/**
 * $Id$
 *
 *<center>
 *               Avocent Corporation. Proprietary Information.
 * \n<em>
 *      This software is supplied under the terms of a license agreement or
 *      nondisclosure agreement with Avocent Corporation, or  one of its 
 *      affiliates, andaffiliates, and may not be used, disseminated, or 
 *      distributed except in accordance with the terms of that agreement.
 *
 *      3541 Gateway Boulevard Fremont, San Jose, CA 94538 U.S.A.
 *\n
 *                  US phone: 510-771-6100
 *
 *        Copyright &copy; 2001-2009 Avocent Corporation. All rights reserved.
 *</em> </center>
 *----------------------------------------------------------------------------\n
 *  MODULES     BIOSPOST TEST PROGRAM                                         \n
 *----------------------------------------------------------------------------\n
 *  @file   aess_dynairqdrv.c
 *  @brief  Dynamic IRQ driver
 *  @internal
 *  $RCSfile: aess_dynairqdrv.c,v $
 *  $Revision: 0.1.0.0 $
 *  $Date: 2009/03/10 13:44:48 $
 *  $Author: Taiyi Kuo $
 *  $Source: $
 *  $Name:  $
 *----------------------------------------------------------------------------*/
/******************************************************************************
* Content
* ----------------
*   aess_dirq_exit()                - The function is invoked when application close the device note.
*   aess_dirq_ioctl()               - The API is invoked when user space application issues an IOCTL command.   
*   aess_dirq_open()                - The function is invoked when application open driver note.
*   aess_dynairq_exit()             - The function is invoked when biospost driver be removed from kernal.
*   aess_dynairq_init()             - The function is invoked when biospost module is loaded into the kernel.
*   aess_dynamic_irq_config()       - To initial the irq(GPIO intrruput type.
*   aess_dynamic_irq_int_clear()    - Clean the interrupt status.
*   aess_register_dynamic_irq_list()- Register the dynamicirq function.
*   GenericISRFunc()                - Generic isr function routing.
*   GPIO_interrupt()                - GPIO interrupt service routing.
*   IRQFuncExec()                   - Registered ISR Function.
*   IRQListGetFree()                - Get an empty point in the array.
*   IRQListSearch()                 - Search the IRQ information in list by device name.
*   IRQListSearchByIRQ()            - Search IRQList Item by IRQ number.
*   IRQListUpdate()                 - Copy an item to sIRQList.
******************************************************************************/
 
 
 
#define AESSDYNAIRQDRV_C
#include <linux/module.h> /* Needed by all modules */
#include <linux/init.h> /* Needed for module_init/exit */
#include <linux/fs.h>
#include <linux/cdev.h>
#include <linux/interrupt.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/bitops.h>
#include "../../eventhandler/eventhandler_driver/aess_eventhandlerdrv.h"
#include "../../gpio/gpio_driver/aess_gpiodrv.h" 
#include "aess_dynairqdrv.h"

//#define DIRQ_DEBUG
/* Debugging Message */
#ifdef DIRQ_DEBUG
#define DEBUG_MSG(fmt, args...) printk( "DynaIRQ: " fmt, ## args)
#else
#define DEBUG_MSG(fmt, args...)
#endif //DIRQ_DEBUG


/* implement debug message with debugfs */
#include <linux/aess_debugfs.h>
static u8   u8tmpcounter = 0;
static u16 u16tmpcounter = 0;
static u32 u32tmpcounter = 0;


/* driver name will passed by insmod command */
static char *driver_name="aess_dynairqdrv"; 

dev_t dirq_dev;
struct cdev *dynairq_cdev;

/*
 ****************************************************************************
 * Dynamic IRQ definitions													*
 ****************************************************************************
 */
static sIRQList stIRQListArray[MAX_DIRQ_LIST];
static int iIRQListCounter=0;

/*
 ****************************************************************************
 * Generic ISR definitions													*
 ****************************************************************************
 */

/* Event handle ID */
#define NAME_GENERIC_ISR	"AESS_GENERIC_ISR"


static sUserModeRecordID stUserModeISRID[32];
static UINT16 cRecUserModeID=0;
static UINT16 u16GISRDriverID=0xffff;


/******************************************************************************
*   FUNCTION        :   IRQListSearch
******************************************************************************/
/**
 *  @brief      Search the IRQ information in list by device name.
 *
 *  @return     Return the point of the find result.
 *              Return NULL if find nothing.                     
 *
 *  @dependency None
 *
 *  @limitation None
 *
 *  @warning    None
 *
 *  @note       None
 *
 *  @internal   Function Type: Internal function \n
 *
 *****************************************************************************/
static sIRQList * IRQListSearch(
                                /** device name */
                                const char * devname
                                )
{
    int i;
    sIRQList *pstIRQList;

    for ( i=0 ; i<iIRQListCounter ; i++ ) 
    {
        pstIRQList = &stIRQListArray[i];

        if (0 == strcmp(pstIRQList->devname, devname)) 
        {
            DRV_INFO("Find IRQ in %d\n", i);
            return pstIRQList;
        }
    }
    return NULL;
}


/******************************************************************************
*   FUNCTION        :   IRQListSearchByIRQ
******************************************************************************/
/**
 *  @brief      Search IRQList Item by IRQ number.
 *
 *  @return     Return the point of the find result.
 *              Return NULL if find nothing.                     
 *
 *  @dependency None
 *
 *  @limitation None
 *
 *  @warning    None
 *
 *  @note       None
 *
 *  @internal   Function Type: Internal function \n
 *
 *****************************************************************************/
static sIRQList * IRQListSearchByIRQ(
                                    /** irq number */    
                                    const int irq
                                    )
{
    int i;
    struct sIrqNum *pstIRQNum, **pp;

    DEBUG_MSG("IRQListSearchByIRQ...\n", i);

    for (i=0 ; i<iIRQListCounter ; i++) 
    {
        pp = &stIRQListArray[i].pstIRQNum;

        while ((pstIRQNum = *pp) != NULL) 
        {
            pp = &pstIRQNum->next;

            if (irq == pstIRQNum->irq_num) 
            {
                DRV_INFO("Find IRQ in %d lists.\n", i);
                return &stIRQListArray[i];
            }
        }
    }

    return NULL;
}


/******************************************************************************
*   FUNCTION        :   IRQListGetFree
******************************************************************************/
/**
 *  @brief      Get an empty point in the array.
 *
 *  @return     Retrun a empty space in the array.
 *
 *  @dependency None
 *
 *  @limitation None
 *
 *  @warning    None
 *
 *  @note       None
 *
 *  @internal   Function Type: Internal function \n
 *
 *****************************************************************************/
static sIRQList * IRQListGetFree(void)
{
    // retrun a empty space in the array
    return &stIRQListArray[iIRQListCounter++];
}

/******************************************************************************
*   FUNCTION        :   IRQListUpdate
******************************************************************************/
/**
 *  @brief      Copy an item to sIRQList.
 *
 *  @return     None.
 *
 *  @dependency None
 *
 *  @limitation None
 *
 *  @warning    None
 *
 *  @note       None
 *
 *  @internal   Function Type: Internal function \n
 *
 *****************************************************************************/
void IRQListUpdate(
                    /** the point of destination struction*/    
                    sIRQList *pstDst, 
                    
                    /** the point of source struction*/    
                    sIRQList *pstSrc
                    )
{
    memcpy(pstDst, pstSrc, sizeof(sIRQList));
}


/******************************************************************************
*   FUNCTION        :   aess_register_dynamic_irq_list
******************************************************************************/
/**
 *  @brief      Register the dynamicirq function.
 *
 *  @return     None.
 *
 *  @dependency None
 *
 *  @limitation None
 *
 *  @warning    None
 *
 *  @note       None
 *
 *  @internal   Function Type: Internal function \n
 *
 *****************************************************************************/
int aess_register_dynamic_irq_list( 
                                    /** device name */    
                                    const char * devname,

                                    /** device id */    
                                    void * dev_id,

                                    /** irqflags */    
                                    unsigned long irqflags,
                                    
                                    /** interrupt service routing */    
                                    irqreturn_t (*handler)(int, void *, struct pt_regs *)
                                    )
{
    sIRQList stIRQList1;
    sIRQList *pstIRQList;

    stIRQList1.pstIRQNum = NULL;
    stIRQList1.devname = devname;
    stIRQList1.dev_id = dev_id;
    stIRQList1.irqflags = irqflags;
    stIRQList1.handler = handler;

    pstIRQList = IRQListSearch(devname);

    if (NULL == pstIRQList) 
    {
        if (iIRQListCounter >= MAX_DIRQ_LIST)
        {
            DRV_ERROR("aess_register_dynamic_irq_list():: The array Overflow!!\n" );
            return -ENOSPC;
        }
        else
        {
            pstIRQList = IRQListGetFree();
        }
    }
    else
    {
        DRV_ERROR("aess_register_dynamic_irq_list():: Same IRQ Name found: Return!\n" );
        return -EEXIST;
    }

    IRQListUpdate(pstIRQList, &stIRQList1);

    return 0;
}

/******************************************************************************
*   FUNCTION        :   aess_dynamic_irq_config
******************************************************************************/
/**
 *  @brief      initial the irq(GPIO intrruput type.
 *
 *  @return     return 0 if success.
 *              return others if fail.
 *          
 *  @dependency None
 *
 *  @limitation None
 *
 *  @warning    None
 *
 *  @note       None
 *
 *  @internal   Function Type: Internal function \n
 *
 *****************************************************************************/
static int aess_dynamic_irq_config(
                                    /** irq number */
                                    int irq, 
                                    
                                    /** interrupt type */
                                    int type, 
                                    
                                    /** enable/disable type */
                                    int enable
                                    )
{
    volatile unsigned char u8IntType;

    if (enable) 
    {
        switch (type) 
        {
        case INTT_RISING_EDGE:
            u8IntType = INPUT_INT_RISING;
            break;

        case INTT_FALLING_EDGE:
            u8IntType = INPUT_INT_FAILING;
            break;

        case INTT_HIGH_LEVEL:
            u8IntType = INPUT_INT_HIGH;
            break;

        case INTT_LOW_LEVEL:
            u8IntType = INPUT_INT_LOW;
            break;
        
        case INTT_BOTH_EDGE:
            u8IntType = INPUT_INT_BOTH;
            break;
        
        default:
            return -1;
        }
    } 
    else 
    {
        u8IntType = INPUT_INT_OFF;
    }

    DRV_INFO("aess_dynamic_irq_config()::PortPin=0x%02x, IntType=0x%02x\n", irq , u8IntType);
    aess_gpio_commander( (irq & 0xff), GPIO_CONFIG, u8IntType, NULL);

    return 0;
}


/******************************************************************************
*   FUNCTION        :   aess_dynamic_irq_int_clear
******************************************************************************/
/**
 *  @brief      Clean the interrupt status.
 *
 *  @return     return 0 if success.
 *              return others if fail.
 *          
 *  @dependency None
 *
 *  @limitation None
 *
 *  @warning    None
 *
 *  @note       None
 *
 *  @internal   Function Type: Internal function \n
 *
 *****************************************************************************/
int aess_dynamic_irq_int_clear(
                                /** irq number */
                                int irq
                               )
{
    UINT8 u8TmpBuf ;

    u8TmpBuf = SET_GPIO_OUTPUT_HIGH;
    aess_gpio_commander( (irq & 0xFF) , GPIO_WRITE, CLR_INT_OUTPUT, (void *)&u8TmpBuf);

    return 0;
}


/******************************************************************************
*   FUNCTION        :   GenericISRFunc
******************************************************************************/
/**
 *  @brief      Dynamiciqr driver interrupt service routing.
 *
 *  @return     return IRQ_NONE if interrupt ocurs.
 *              return others if fail.
 *          
 *  @dependency None
 *
 *  @limitation None
 *
 *  @warning    None
 *
 *  @note       None
 *
 *  @internal   Function Type: Internal function \n
 *
 *****************************************************************************/
static irqreturn_t GenericISRFunc(
                                    /** irq number */
                                    int irq, 
                                    
                                    /** device id */
                                    void *dev_id, 
                                    
                                    /** point of pt_regs */
                                    struct pt_regs *regs
                                    )
{
    volatile int i;

    DEBUG_MSG("GenericISRFunc(), irq=%d\n",irq);

    if (0xffff == u16GISRDriverID) 
    {
        DRV_ERROR("Interrupt arrive!!::GeneIRQ: Event handle ID no initial\n");
        return IRQ_NONE;
    }

    if (0 == cRecUserModeID) 
    {
        DRV_ERROR("Interrupt arrive!!::GeneIRQ: No record.\n");
        return IRQ_NONE;
    }

    for (i=0; i<cRecUserModeID ; i++) 
    {
        if (irq == stUserModeISRID[i].u16NumIRQ) 
        {
            aess_generate_driver_event(u16GISRDriverID, stUserModeISRID[i].u32EventID);
            return IRQ_HANDLED;
        }
    }

    DRV_ERROR("Interrupt arrive!!::GeneIRQ: IRQ%d record not found !!\n", irq);

    return IRQ_NONE;
}

/******************************************************************************
*   FUNCTION        :   aess_dirq_open
******************************************************************************/
/**
 *  @brief      The function is invoked when application open driver note.
 *
 *  @return     STATUS_OK never fail.                   
 *
 *  @dependency None
 *
 *  @limitation None
 *
 *  @warning    None
 *
 *  @note       None
 *
 *  @internal   Function Type: Global function \n
 *
 *****************************************************************************/
static int aess_dirq_open(
                             /** inode structure point */
                             struct inode *inode, 

                             /** file structure point */
                             struct file *filp
                            )
{
    static uint uopen = 0;

    if (0 == uopen)
    {
        uopen = 1;
        DRV_INFO("aess_dirq_open()::");
    }
    else
        DRV_INFO("aess_dirq_open()::Already Opend!!");

    return 0;
}

/******************************************************************************
*   FUNCTION        :   aess_dirq_exit
******************************************************************************/
/**
 *  @brief      The function is invoked when application close the device note.
 *
 *  @return     STATUS_OK never fail.                   
 *
 *  @dependency None
 *
 *  @limitation None
 *
 *  @warning    None
 *
 *  @note       None
 *
 *  @internal   Function Type: Global function \n
 *
 *****************************************************************************/
int aess_dirq_exit(
                             /** inode structure point */
                             struct inode *inode, 

                             /** file structure point */
                             struct file *filp
                    )
{
    DRV_INFO("aess_dirq_exit()::\n");
    return 0;
}

/******************************************************************************
*   FUNCTION        :   aess_dirq_ioctl
******************************************************************************/
/**
 *  @brief      The API is invoked when user space application issues an IOCTL command. 
 *              The following is the list of commands that supported currently.
 *              1.	AESS_IOCTL_DRIVER_INIT
 *                  Initialize Linux BIOS Post Codes driver according to the given parameters.
 *              2.	AESS_IOCTL_DYNAIRQ_INIT
 *                  Copy data from driver to user space application.
 *              3.	AESS_IOCTL_DYNAIRQ_CONFIG
 *                  Reset index and size of a data buffer to zero.
 *              4.	AESS_IOCTL_DYNAIRQ_CLEAR
 *                  Reset index and size of a data buffer to zero.
 *              5.	AESS_IOCTL_GENEISR_INIT
 *                  Reset index and size of a data buffer to zero.
 *              6.	AESS_IOCTL_USERMODE_ISRID
 *                  Reset index and size of a data buffer to zero.
 *
 *  @return     STATUS_OK if driver execute the command success.
 *              Others if execute the command fail.                   
 *
 *  @dependency None
 *
 *  @limitation None
 *
 *  @warning    None
 *
 *  @note       None
 *
 *  @internal   Function Type: Global function \n
 *
 *****************************************************************************/
static int aess_dirq_ioctl(
                              /** inode structure point */
                              struct inode *inode, 
           
                              /** file structure point */
                              struct file *filp,
           
                              /** command from user space */
                              unsigned int cmd, 
    
                              /** sIRQInfoType structure point */
                              unsigned long arg
)
{
    sIRQInfoType *pstDIRQ = NULL;
    sIRQList *pstIRQList= NULL;
    uint irq_num = 0;
    struct sIrqNum *pstIRQNum, **pp;
    int i,findUMISR;
    int int_type, int_enable;
    UINT8 u8tmp;

    if ((cmd != AESS_IOCTL_DRIVER_INIT) && (0 == arg)) 
    {
        DRV_ERROR("aess_dirq_ioctl()::(cmd != AESS_IOCTL_DRIVER_INIT) && (0 == arg)\n");
        return -EFAULT;
    }

    /* Check the DynaIRQ structure address is OK */
    if ( cmd == AESS_IOCTL_DYNAIRQ_INIT || cmd == AESS_IOCTL_DYNAIRQ_CONFIG || cmd == AESS_IOCTL_DYNAIRQ_CLEAR ) 
    {
        pstDIRQ = (sIRQInfoType *) arg;
        irq_num = pstDIRQ->param1.u16IRQNum;
        
        if(!access_ok(VERIFY_READ, (void *) pstDIRQ, sizeof(sIRQInfoType)))
        {
            DRV_INFO("aess_dirq_ioctl()::struct access error!!\n");
            /* error occur */
            return -EFAULT;
        }
        /* Check input params */
        if ( (cmd == AESS_IOCTL_DYNAIRQ_INIT) && (NULL == pstDIRQ->pISRName) )
        {
            DRV_ERROR("aess_dirq_ioctl()::Check input params\n");
            return -EFAULT;
        }

        if( !(irq_num > 0 || irq_num < GPIO_MAX_PIN_NUM ) ) 
        {
            DRV_ERROR("aess_dirq_ioctl()::Check input params\n");
            return -EFAULT;
        }
    }

    DRV_INFO("aess_dirq_ioctl():: cmd = %d \n" ,cmd );

    switch (cmd) 
    {
        case AESS_IOCTL_DRIVER_INIT:

            /* free register IRQ */
            findUMISR = 0;

            for (i=0 ; i<iIRQListCounter ; i++) 
            {
                pp = &stIRQListArray[i].pstIRQNum;

                if (strcmp(stIRQListArray[i].devname,NAME_GENERIC_ISR)==0) 
                {
                    findUMISR = 1;
                }

                while ((pstIRQNum = *pp) != NULL) 
                {
                    pp = &pstIRQNum->next;
                    kfree(pstIRQNum);
                }
                stIRQListArray[i].pstIRQNum = NULL;
            }
            /* initial global var */
            iIRQListCounter = findUMISR;
            cRecUserModeID = 0;
            u16GISRDriverID = 0xffff;

            break;

        case AESS_IOCTL_DYNAIRQ_INIT:

            DRV_INFO("aess_dirq_ioctl(AESS_IOCTL_DYNAIRQ_INIT)::ISR name = %s : RQ number = %d\n", pstDIRQ->pISRName, irq_num );

            /* Search IRQList by input name */
            pstIRQList = IRQListSearch(pstDIRQ->pISRName);

            if (NULL == pstIRQList) 
            {
                DRV_ERROR("aess_dirq_ioctl()::No such IRQ List\n");
                return -ENXIO;
            }

            /* call GPIO API to setup ext-IRQ */
            int_type = (pstDIRQ->param3 >> 5) & 0x7;
            int_enable = (pstDIRQ->param3 >> 8) & 0x1;

            if (aess_dynamic_irq_config(pstDIRQ->param1.u16IRQNum , int_type, int_enable) < 0)
            {
                DRV_ERROR("aess_dirq_ioctl()::aess_dynamic_irq_config() return fail\n");
                return -EFAULT;
            }

            // make suer this irq_no nerver be registered
            if (IRQListSearchByIRQ(irq_num)) 
            {
                DRV_ERROR("aess_dirq_ioctl()::aess_dynamic_irq_config() return fail\n");
                return -EBUSY;
            }
            // Save IRQ into pstDIRQ 
            pp = &pstIRQList->pstIRQNum;

            while ((pstIRQNum = *pp) != NULL) 
            {
                pp = &pstIRQNum->next;
            }

            /* alloc new IRQ item */
            pstIRQNum = kmalloc(sizeof(struct sIrqNum), GFP_KERNEL);

            if (!pstIRQNum)
            {
                DRV_ERROR("aess_dirq_ioctl():: memory access fail!!\n");
                return -ENOMEM;
            }   
            pstIRQNum->irq_num = irq_num;
            pstIRQNum->next = NULL;
            *pp = pstIRQNum;

            break;

        case AESS_IOCTL_DYNAIRQ_CONFIG:
            
            /* call GPIO API to setup ext-IRQ */
            int_type = (pstDIRQ->param3 >> 5) & 0x7;
            int_enable = (pstDIRQ->param3 >> 8) & 0x1;
            
            DRV_INFO("aess_dirq_ioctl(AESS_IOCTL_DYNAIRQ_INIT)::pstDIRQ->param1.u16IRQNum=%d  , int_type=%d , int_enable=%d \n ",
                        pstDIRQ->param1.u16IRQNum ,int_type ,int_enable );
            
            if (aess_dynamic_irq_config(pstDIRQ->param1.u16IRQNum , int_type,   int_enable) < 0)
            {
                DRV_ERROR("aess_dirq_ioctl():: Enable interrupt fail(AESS_IOCTL_DYNAIRQ_CONFIG)!!\n");
                return -EFAULT;
            }
            break;

		case AESS_IOCTL_DYNAIRQ_CLEAR:

            /* Clear GIPO ISR */
            u8tmp = 1 ;
            if ( aess_gpio_commander(pstDIRQ->param1.u16IRQNum, GPIO_WRITE, CLR_INT_OUTPUT, (void *)&u8tmp) < 0)
            {			
                DRV_ERROR("aess_dirq_ioctl():: Enable interrupt fail(AESS_IOCTL_DYNAIRQ_CLEAR)!!\n");
                return -EFAULT;
            }
            break;

		case AESS_IOCTL_GENEISR_INIT:

            /* Set DriverID for event handler */
            if( copy_from_user(&u16GISRDriverID, (UINT16 *)arg, sizeof(UINT16))  )
            {
                DRV_ERROR("copy_from_user() Fault!!");
                return -EFAULT ;
            }

            break;

		case AESS_IOCTL_USERMODE_ISRID:

            /* Set User Mode ISR ID for event handler */
            if (cRecUserModeID >= 32)
                return -EFAULT;
            
            if( copy_from_user(&stUserModeISRID[cRecUserModeID], (sUserModeRecordID *)arg, sizeof(stUserModeISRID)) )
            {
                DRV_ERROR("copy_from_user() Fault!!");
                return -EFAULT ;
            }
            
            DRV_INFO("aess_dirq_ioctl(AESS_IOCTL_USERMODE_ISRID):: IRq=%d , EventID=%d\n"   
                ,stUserModeISRID[cRecUserModeID].u16NumIRQ ,stUserModeISRID[cRecUserModeID].u32EventID);            

            cRecUserModeID++;

            break;

        default:
            DRV_ERROR("aess_dirq_ioctl():: command error\n");
            return -EINVAL;
	}
	  
	return 0;
}

/******************************************************************************
*   FUNCTION        :   IRQFuncExec
******************************************************************************/
/**
 *  @brief      Registered ISR Function.
 *
 *  @return     Return 0 if execute the register function success.
 *              Return -1 if execute the register function success.                   
 *
 *  @dependency None
 *
 *  @limitation None
 *
 *  @warning    None
 *
 *  @note       None
 *
 *  @internal   Function Type: Internal function \n
 *
 *****************************************************************************/ 
static inline int IRQFuncExec(
                                /** irq number */
                                int irq
                              )
{
    volatile sIRQList *pstIRQList;

    pstIRQList = IRQListSearchByIRQ(irq);
    if (NULL == pstIRQList || NULL == pstIRQList->handler) 
    {
        return -1;
    }

    /* call handler function */
    (pstIRQList->handler)(irq, pstIRQList->dev_id, NULL);

    return 0;
}

/******************************************************************************
*   FUNCTION        :   GPIO_interrupt
******************************************************************************/
/**
 *  @brief      GPIO interrupt service routing. 
 *
 *  @return     IRQ_NONE if the interrupt not occured biospost .
 *              IRQ_HANDLED if biospsot interrupt occured.                   
 *
 *  @dependency None
 *
 *  @limitation None
 *
 *  @warning    None
 *
 *  @note       None
 *
 *  @internal   Function Type: Internal function \n
 *
 *****************************************************************************/
static irqreturn_t GPIO_interrupt(
                                    /** irq number */
                                    int irq_share, 
                                    
                                    /** device id */
                                    void *dev_id
                                    )
{
    volatile int GPIO_Event_Pin ;
    volatile UINT8  u8TmpBuf = 1;  
    volatile int i = 0;
    volatile int j = 0;
    
    volatile UINT32 IntValue[2] ;

    if( aess_gpio_GetEventPin( (UINT32*)&(IntValue[0]) ,(UINT32*)&(IntValue[1]) ) !=0 )
    {

        DRV_INFO("GPIO_interrupt()::IntValue=0x%08x,IntValueExt=0x%08x\n",IntValue[0] ,IntValue[1] );    
        for( j = 0 ; j < 2 ; j++ )
        {
            for( i = 0 ; i < 32 ; i ++ )
            {
                if( ( (IntValue[j]>> i) & 0x00000001 ) == 1 )
                {
                    GPIO_Event_Pin = i + j*32 ;
                    // clean the interrupt interrupt status
                    u8TmpBuf = 1;
                    aess_gpio_commander( GPIO_Event_Pin , GPIO_WRITE, CLR_INT_OUTPUT, (void *)&u8TmpBuf);

                    // clear buf
                    u8TmpBuf = 0 ; 

                    // double the specific GPIO pin did enable the interrupt?
                    aess_gpio_commander( GPIO_Event_Pin , GPIO_READ, INT_ENABLE_INPUT, (void *)&u8TmpBuf);

                    // disable irq 
                    aess_dynamic_irq_config( GPIO_Event_Pin ,0 , 0 );
                    
                    if(  u8TmpBuf != 0 ) // interrupt enabled
                    {
                        if ( IRQFuncExec( GPIO_Event_Pin ) < 0) 
                        {
                            DRV_ERROR("Call IRQFuncExec() and return fail!! EventPin=%d\n",GPIO_Event_Pin);
                        }
                    }            
                }
            }                
        }
        return IRQ_HANDLED;    

    }
    DRV_ERROR("GPIO interrupt occurred but pin number ??!\n");
    return IRQ_NONE;    
}


static struct file_operations aess_dirq_fops = 
{
    .open = aess_dirq_open, 
    .ioctl = aess_dirq_ioctl,
    .release = aess_dirq_exit, 
};


/******************************************************************************
*   FUNCTION        :   aess_dynairq_init
******************************************************************************/
/**
 *  @brief      The function is invoked when biospost module is loaded into the kernel.
 *
 *  @return     STATUS_OK if driver initial success.
 *              Others if initial fail.                   
 *
 *  @dependency None
 *
 *  @limitation None
 *
 *  @warning    None
 *
 *  @note       None
 *
 *  @internal   Function Type: Global function  \n
 *
 *****************************************************************************/
static int __init aess_dynairq_init(void)
{
    dynairq_cdev = cdev_alloc();
    cdev_init(dynairq_cdev, &aess_dirq_fops);

    DEBUG_MSG("aess_dynairq_init\n");

    if ( alloc_chrdev_region(&dirq_dev, 0, 1, driver_name) < 0) 
    {
        printk (KERN_ERR "DynaIRQ:Registering the Dynamic IRQ character device failed!!\n");
        return -EFAULT;
    }

    printk("mknod /dev/aess_dynairqdrv c %d 0\n", MAJOR(dirq_dev));

    // register thre dynamic_irq driver    
    cdev_add(dynairq_cdev, dirq_dev, 1);

    if ( request_irq( INT_GPIO , GPIO_interrupt ,  IRQF_SAMPLE_RANDOM | IRQF_DISABLED,"GPIO_INTERRUT" , NULL  ) ) 
    {
        printk ( KERN_ERR "DynaIRQ::Can;t register the GPIO IRQ !!");
        return -EFAULT;
    }
    
    // register the generic isr and the their call back function to GenericISRFunc
    if( aess_register_dynamic_irq_list( NAME_GENERIC_ISR, NULL, 0 , (void *)GenericISRFunc) < 0 )
    {
        printk (KERN_ERR "DynaIRQ:Registering the Dynamic IRQ list failed !\n");
        free_irq( INT_GPIO, NULL );
        return -EFAULT ;
    }
    
    /* initial debugfs related resource */
    aess_debugfs_default_create(driver_name,NULL,0,0);
    aess_debugfs_create_file(driver_name,"u32",(u32 *)&u32tmpcounter,DBG_TYPE32,0);
    aess_debugfs_create_file(driver_name,"u16",(u16 *)&u16tmpcounter,DBG_TYPE16,0);
    aess_debugfs_create_file(driver_name,"u8",(u8 *)&u8tmpcounter,DBG_TYPE8,0);	

    DRV_INFO("aess_gpio_init()::return (STATUS_OK)");    

    return 0;
}

/******************************************************************************
*   FUNCTION        :   aess_dynairq_exit
******************************************************************************/
/**
 *  @brief      The function is invoked when biospost driver be removed from kernal.
 *
 *  @return     None.                   
 *
 *  @dependency None
 *
 *  @limitation None
 *
 *  @warning    None
 *
 *  @note       None
 *
 *  @internal   Function Type: Global function \n
 *
 *****************************************************************************/
static void __exit aess_dynairq_exit(void)
{
    DRV_INFO("aess_dynairq_exit()\n");

    free_irq( INT_GPIO, NULL );

    cdev_del(dynairq_cdev);
    unregister_chrdev_region(dirq_dev, 1);

    /* remove debugfs related resource */
    aess_debugfs_remove(driver_name,NULL,0,0); 	    
}

MODULE_DESCRIPTION("AESS Dynamic IRQ Driver");
MODULE_AUTHOR("Taiyi Kuo<taiti.kuo@avocent.com>");
MODULE_LICENSE("GPL");
MODULE_VERSION("0.1.0.0");
module_init(aess_dynairq_init);
module_exit(aess_dynairq_exit);
module_param(driver_name, charp, S_IRUGO);

EXPORT_SYMBOL(aess_register_dynamic_irq_list);
EXPORT_SYMBOL(aess_dynamic_irq_int_clear);



/* End of code */
