/*
 *  linux/arch/arm/mach-umvp/irq.c
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <linux/init.h>
#include <linux/stddef.h>
#include <linux/list.h>
#include <linux/timer.h>

#include <asm/hardware.h>
#include <asm/irq.h>
#include <asm/io.h>

#include <asm/mach/irq.h>
#include <asm/arch/vic.h>

#define VIC_BASE_VA             IO_ADDRESS(ASPEED_VIC_BASE) 

#define UMVP_READ_REG(r)        (*((volatile unsigned int *) (r)))
#define UMVP_WRITE_REG(r,v)     (*((volatile unsigned int *) (r)) = ((unsigned int)   (v)))

static void vic_reset(void)
{
    UMVP_WRITE_REG(VIC_BASE_VA + ASPEED_VIC_INT_SEL_OFFSET,      0x00000000);
    UMVP_WRITE_REG(VIC_BASE_VA + ASPEED_VIC_ENABLE_SET_OFFSET,   0x00000000);
    UMVP_WRITE_REG(VIC_BASE_VA + ASPEED_VIC_ENABLE_CLEAR_OFFSET, 0xFFFFFFFF);
    UMVP_WRITE_REG(VIC_BASE_VA + ASPEED_VIC_EDGE_CLEAR_OFFSET,   0xFFFFFFFF);

}

void vic_set_intr_trigger(int irqnr, vic_intr_trig_mode_E mode)
{
    unsigned int nMask;
    unsigned int nRegValue;
    unsigned int nVICVA;
    unsigned int nVIC_SENSE_OFFSET, nVIC_EVENT_OFFSET, nVIC_BOTH_EDGE_OFFSET;

    {
        nMask = (1 << irqnr);
	nVICVA = VIC_BASE_VA;		/* switch to VIC1 */
	nVIC_SENSE_OFFSET = ASPEED_VIC_SENSE_OFFSET;
	nVIC_EVENT_OFFSET = ASPEED_VIC_EVENT_OFFSET;
	nVIC_BOTH_EDGE_OFFSET = ASPEED_VIC_BOTH_EDGE_OFFSET;	
    }

    /* edge or level trigger */
    nRegValue = UMVP_READ_REG(nVICVA + nVIC_SENSE_OFFSET);
    if (mode == vicc_edge_activeFalling || mode == vicc_edge_activeRaising ||
        mode == vicc_edge_activeBoth)
    {
        /* edge trigger : corresponding bit shall be reset as 0 */
        nRegValue &= (~nMask);
    }
    else
    {
        /* level trigger : corresponding bit shall be set as 1 */
        nRegValue |= nMask;
    }
    UMVP_WRITE_REG(nVICVA + nVIC_SENSE_OFFSET, nRegValue);



    /* active high or active low */
    nRegValue = UMVP_READ_REG(nVICVA + nVIC_EVENT_OFFSET);
    if (mode == vicc_edge_activeFalling || mode == vicc_level_activeLow)
    {
        /* active low : corresponding bit shall be reset as 0 */
        nRegValue &= (~nMask);
    }
    else if (mode == vicc_edge_activeRaising || mode == vicc_level_activeHigh)
    {
        /* active high : corresponding bit shall be set as 1 */
        nRegValue |= nMask;
    }
    UMVP_WRITE_REG(nVICVA + nVIC_EVENT_OFFSET, nRegValue);



    /* both edge? */
    nRegValue = UMVP_READ_REG(nVICVA + nVIC_BOTH_EDGE_OFFSET);
    if (mode == vicc_edge_activeBoth)
    {
        /* both edge : corresponding bit shall be set as 1 */
        nRegValue |= nMask;
    }
    else
    {
        /* single edge : corresponding bit shall be reset as 0 */
        nRegValue &= (~nMask);
    }
    UMVP_WRITE_REG(nVICVA + nVIC_BOTH_EDGE_OFFSET, nRegValue);
}

static void vic_enable_intr(int irq)
{
	register unsigned int regVal;

	{
		regVal = UMVP_READ_REG(VIC_BASE_VA + ASPEED_VIC_ENABLE_SET_OFFSET);
		regVal |= (1 << irq);
		UMVP_WRITE_REG(VIC_BASE_VA + ASPEED_VIC_ENABLE_SET_OFFSET, regVal);
	}	

}

static void vic_disable_intr(int irq)
{
	register unsigned int regVal;

	{
		regVal = UMVP_READ_REG(VIC_BASE_VA + ASPEED_VIC_ENABLE_CLEAR_OFFSET);
		regVal |= (1 << irq);
		UMVP_WRITE_REG(VIC_BASE_VA + ASPEED_VIC_ENABLE_CLEAR_OFFSET, regVal);
	}	

}

void vic_clear_intr(int irq)
{
	{
		UMVP_WRITE_REG(VIC_BASE_VA + ASPEED_VIC_EDGE_CLEAR_OFFSET, (1 << irq));
	}

}

static void sc_mask_irq(unsigned int irq)
{
	vic_disable_intr(irq);
}

static void sc_unmask_irq(unsigned int irq)
{
	vic_enable_intr(irq);
}
 
static struct irqchip sc_chip = {
	.ack	= sc_mask_irq,
	.mask	= sc_mask_irq,
	.unmask = sc_unmask_irq,
}; 
 
void __init aspeed_init_irq(void)
{
	unsigned int i;

	vic_reset();

	for (i = 0; i < NR_IRQS; i++) 
	{
		{
			/* apply VIC1 mask (VIC means VIC1) */
			if ((1 << i) & ASPEED_VIC_VALID_INTMASK) 
			{
				set_irq_chip(i, &sc_chip);
				set_irq_handler(i, do_level_IRQ);
				set_irq_flags(i, IRQF_VALID);
			}
		}

	}
        
	/* enable INT_VIC2IRQ on VIC1 for VIC2 be able to launch interrupt */
	/* vic_set_intr_trigger(INT_VIC2IRQ, vicc_level_activeLow);  */


	/* note that : VIC1 needs to be programmed to launch IRQ to CPU for INT_VIC2IRQ */
	/*  and this has been done in the vic_reset(). */

	/* Enable interrupt on VIC1 for INT_VIC2IRQ. */
	/* Please note that you can't use "enable_irq()" to enable this interrupt */
	/* since the program will not generate this interrupt. Please refer to */
	/* "entry-macro.S". */
	/* vic_enable_intr(INT_VIC2IRQ); */
}

