/*
 * linux/drivers/usb/gadget/aspeed_udc.c
 * ASpeed 2050 on-chip full speed USB device controllers
 *
 * Copyright (C) 2008, Avocent Corporation
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#ifndef LINUX_VERSION_CODE
  #include <linux/version.h>
#endif

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/delay.h>
#include <linux/list.h>
#include <linux/interrupt.h>
#include <linux/proc_fs.h>
#include <linux/device.h>
#include <linux/workqueue.h>

#include <asm/io.h>
#include <asm/system.h>
#include <asm/hardware.h>

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,21)
  #include <linux/usb_ch9.h>
#else
  #include <linux/usb/ch9.h>
#endif

#include <linux/usb_gadget.h>
#include <linux/platform_device.h>
#include <asm/cacheflush.h>
#include <linux/dma-mapping.h>

#include "ast2050_udc.h"

#undef	DEBUG

/*
 * debug routines
 */
#ifndef	DEBUG
#ifndef DEBUG_IRQ
# define DEBUG_IRQ(fmt,args...)
#endif
#ifndef DEBUG_EP0
# define DEBUG_EP0(fmt,args...)
#endif
#ifndef DEBUG_SETUP
# define DEBUG_SETUP(fmt,args...)
#endif
#ifndef DEBUG_ENTRY
# define DEBUG_ENTRY(fmt,args...)
#endif
#ifndef DEBUG_LIST
# define DEBUG_LIST(fmt,args...)
#endif
#ifndef DEBUG_RING
# define DEBUG_RING(fmt,args...)
#endif
#ifndef DEBUG_PRINT
# define DEBUG_PRINT(fmt,args...)
#endif
#endif

/*
 * debug state
 */
static unsigned long dbg = 0;

module_param(dbg, ulong, 0);
MODULE_PARM_DESC(dbg, "debug state");

#define	DEBUG_PRINT_ENABLE	1
#define	DEBUG_SETUP_ENABLE	2
#define	DEBUG_IRQ_ENABLE	4
#define	DEBUG_EP0_ENABLE	8
#define	DEBUG_LIST_ENABLE 	16
#define	DEBUG_ENTRY_ENABLE	32
#define	DEBUG_RING_ENABLE	64	

#define	DRIVER_DESC		"ASPEED 2050 USB Device Controller"
#define	DRIVER_VERSION		__DATE__

static const char driver_name[] = "g_ast2050_udc";
static const char driver_desc[] = DRIVER_DESC;
static const char ep0name[] = "ep0-control";

static struct aspeed_udc *the_controller = NULL;

#define	LINESIZE	L1_CACHE_BYTES
#define	LINEMASK	0xFFFFFFE0

#define HID_REQ_SET_REPORT	0x9
#define STALL_SUPPORT       1

/*
 * local definitions
 */
int device_address;
int device_address_change = 0;

/*
 * debug strings
 */
char * typeName[] = {
	"standard",
	"class",
	"vendor",
	"reserved" 
};

char * recipientName[] = {
	"device",
	"interface",
	"endpoint",
	"other",
	"port",
	"rpipe"
};

char * requestName[] = {
	"get status",
	"clear feature",
	"?",
	"set feature",
	"?",
	"set address",
	"get descriptor",
	"set descriptor",
	"get configuration",
	"set configuration",
	"get interface",
	"set interface",
	"synch frame",
	"set encryption",
	"get encryption",
	"set handshake",
	"get handshake",
	"set connection",
	"set security data",
	"get security data",
	"set wusb data",
	"loopback data write",
	"loopback data read",
	"set interface ds"
};
	
/*
 * local declarations
 */
#ifdef	DEBUG
/* debug routines */
static void DEBUG_PRINT(char *, ...);
static void DEBUG_SETUP(char *, ...);
static void DEBUG_IRQ(char *, ...);
static void DEBUG_EP0(char *, ...);
static void DEBUG_LIST(char *, ...);
static void DEBUG_ENTRY(char *, ...);
#endif

/* gadget api routines */
int usb_gadget_register_driver(struct usb_gadget_driver *);
int usb_gadget_unregister_driver(struct usb_gadget_driver *);
static int aspeed_ep_enable(struct usb_ep *,
	const struct usb_endpoint_descriptor *);
static int aspeed_ep_disable(struct usb_ep *);
static struct usb_request *aspeed_alloc_request(struct usb_ep *, gfp_t);
static void aspeed_free_request(struct usb_ep *, struct usb_request *);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,23)
static void *aspeed_alloc_buffer(struct usb_ep *, unsigned, dma_addr_t *,
				  gfp_t);
static void aspeed_free_buffer(struct usb_ep *, void *, dma_addr_t,
	unsigned);
#endif
static int aspeed_queue(struct usb_ep *, struct usb_request *, gfp_t);
static int aspeed_dequeue(struct usb_ep *, struct usb_request *);
static int aspeed_set_halt(struct usb_ep *, int);
static int aspeed_fifo_status(struct usb_ep *);
static void aspeed_fifo_flush(struct usb_ep *);

/* standard udc routines */
static void udc_disable(struct aspeed_udc *);
static void udc_enable(struct aspeed_udc *);
static void udc_reinit(struct aspeed_udc *);
static void udc_ep0_kick(struct aspeed_udc *, struct aspeed_ep *);
static int udc_ep0_in(struct aspeed_udc *, struct aspeed_ep *);
static void udc_ep0_out(struct aspeed_udc *, struct aspeed_ep *);
static void udc_done(struct aspeed_ep *, struct aspeed_request *, int);
static void udc_nuke(struct aspeed_ep *, int);
static void udc_flush(struct aspeed_ep *);
static void udc_stop_activity(struct aspeed_udc *, struct usb_gadget_driver *);

/* fifo routines */
static int write_fifo(struct aspeed_ep *ep, struct aspeed_request *req, int);
static int write_fifo_ep0(struct aspeed_ep *ep, struct aspeed_request *req);
static int read_fifo(struct aspeed_ep *ep, struct aspeed_request *req);
static int read_fifo_ep0(struct aspeed_ep *ep, struct aspeed_request *req);

/* EP0 routines */
static int aspeed_handle_get_status(struct aspeed_udc *,
				     struct usb_ctrlrequest *);
static void aspeed_ep0_setup(struct aspeed_udc *);

/* low-level aspeed 2050 usbcore routines */
static void read_setup_packet(struct aspeed_udc *, struct usb_ctrlrequest *);
static void disable_interrupts(void);
static void enable_interrupts(void);
static void set_address(struct aspeed_ep *, u8);
static int enable_endpoint(struct aspeed_ep *);
static void disable_endpoint(struct aspeed_ep *);
static int get_speed(void);
#ifdef	STALL_SUPPORT
static void reset_data_toggle(struct aspeed_ep *);
static int set_stall(struct aspeed_ep *);
static int clear_stall(struct aspeed_ep *);
static int check_stall(struct aspeed_ep *);
#endif
static void idle_endpoint(struct aspeed_ep *);

/*
 * low-level aspeed 2050 usbcore transfer routines
 * most routines are endpoint type specific
 */
static int recv_out_packet_control(struct aspeed_ep *, u8 *, int);
static int recv_out_packet_int(struct aspeed_ep *, u8 *, int);
static int recv_out_packet_bulk(struct aspeed_ep *, u8 *, int);
static int data_in_fifo_control(struct aspeed_ep *);
static int data_in_fifo_int(struct aspeed_ep *);
static int data_in_fifo_bulk(struct aspeed_ep *);
static int transmit_in_packet_control(struct aspeed_ep *, u8 *, int);
static int transmit_in_packet_int(struct aspeed_ep *, u8 *, int);
static int transmit_in_packet_bulk(struct aspeed_ep *, u8 *, int);
static void terminate_in_packet_control(struct aspeed_ep *);
static void terminate_in_packet_int(struct aspeed_ep *);
static void terminate_in_packet_bulk(struct aspeed_ep *);
static void ready_to_receive_control(struct aspeed_ep *);
static void ready_to_receive_int(struct aspeed_ep *);
static void ready_to_receive_bulk(struct aspeed_ep *);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,23)
static void timeout_transfer(void *);
#else
static void timeout_transfer(struct work_struct *);
#endif

/* interrupt routines */
static void aspeed_in_epn(struct aspeed_ep *);
static void aspeed_out_epn(struct aspeed_ep *);
static void aspeed_reset_intr(struct aspeed_udc *);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,23)
static irqreturn_t aspeed_usb_interrupt (int, void *, struct pt_regs *);
#else
static irqreturn_t aspeed_usb_interrupt (int, void *);
#endif

static void usbcore_reset(void);
static void usbcore_start(void);
static void usbcore_stop(void);
static void usbcore_clock_start(void);
static void usbcore_clock_stop(void);

/*
 * endpoint type specific transfer routines
 */
static struct transfer_ops transfer_ops_control = {
	recv_out_packet_control,
	data_in_fifo_control,
	transmit_in_packet_control,
	terminate_in_packet_control,
	ready_to_receive_control,
};

static struct transfer_ops transfer_ops_int = {
	recv_out_packet_int,
	data_in_fifo_int,
	transmit_in_packet_int,
	terminate_in_packet_int,
	ready_to_receive_int,
};

static struct transfer_ops transfer_ops_bulk = {
	recv_out_packet_bulk,
	data_in_fifo_bulk,
	transmit_in_packet_bulk,
	terminate_in_packet_bulk,
	ready_to_receive_bulk,
};

/*
 * gadget api
 */
static struct usb_ep_ops aspeed_ep_ops = {
	.enable = aspeed_ep_enable,
	.disable = aspeed_ep_disable,

	.alloc_request = aspeed_alloc_request,
	.free_request = aspeed_free_request,

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,23)
	.alloc_buffer = aspeed_alloc_buffer,
	.free_buffer = aspeed_free_buffer,
#endif

	.queue = aspeed_queue,
	.dequeue = aspeed_dequeue,

	.set_halt = aspeed_set_halt,
	.fifo_status = aspeed_fifo_status,
	.fifo_flush = aspeed_fifo_flush,
};

#ifdef	DEBUG
/*
 * debug routines
 */
static void DEBUG_IRQ(char *fmt, ...)
{
	va_list args;
	char buf[100];

	if (dbg & DEBUG_IRQ_ENABLE) {
		va_start(args, fmt);
		vsnprintf(buf, 100, fmt, args);
		va_end(args);
		printk(buf);
	}
}

static void DEBUG_EP0(char *fmt, ...)
{
	va_list args;
	char buf[100];

	if (dbg & DEBUG_EP0_ENABLE) {
		va_start(args, fmt);
		vsnprintf(buf, 100, fmt, args);
		va_end(args);
		printk(buf);
	}
}

static void DEBUG_SETUP(char *fmt, ...)
{
	va_list args;
	char buf[100];

	if (dbg & DEBUG_SETUP_ENABLE) {
		va_start(args, fmt);
		vsnprintf(buf, 100, fmt, args);
		va_end(args);
		printk(buf);
	}
}

static void DEBUG_ENTRY(char *fmt, ...)
{
	va_list args;
	char buf[100];

	if (dbg & DEBUG_ENTRY_ENABLE) {
		va_start(args, fmt);
		vsnprintf(buf, 100, fmt, args);
		va_end(args);
		printk(buf);
	}
}

static void DEBUG_LIST(char *fmt, ...)
{
	va_list args;
	char buf[100];

	if (dbg & DEBUG_LIST_ENABLE) {
		va_start(args, fmt);
		vsnprintf(buf, 100, fmt, args);
		va_end(args);
		printk(buf);
	}
}

static void DEBUG_PRINT(char *fmt, ...)
{
	va_list args;
	char buf[100];

	if (dbg & DEBUG_PRINT_ENABLE) {
		va_start(args, fmt);
		vsnprintf(buf, 100, fmt, args);
		va_end(args);
		printk(buf);
	}
}
#endif

/*
 * proc routines
 */
#ifdef CONFIG_USB_GADGET_DEBUG_FILES

static const char proc_node_name[] = "driver/udc";

/**
 * show_usbcore_register - display usbcore register in /proc
 * @next:
 * @size:
 * @fmt: format string
 * @offset: register offset
 */
static int show_usbcore_register(char *next, unsigned size, char *fmt,
	u32 offset)
{
	return scnprintf(next, size, fmt, ioread32(IO_ADDRESS(offset)));
}

/**
 * show_usbcore_register_array - display 4 usbcore registers in /proc
 * @next:
 * @size:
 * @fmt: format string
 * @offset: register offset base
 * @asize: array element size
 */
static int show_usbcore_register_array(char *next, unsigned size, char *fmt,
	u32 offset, int asize)
{
	return scnprintf(next, size, fmt,
		ioread32(IO_ADDRESS(offset)),
		ioread32(IO_ADDRESS(offset+asize)),
		ioread32(IO_ADDRESS(offset+(2*asize))),
		ioread32(IO_ADDRESS(offset+(3*asize))));
}

/**
 * udc_proc_read - /proc routine, show status
 * @page:
 * @start:
 * @off:
 * @count:
 * @eof:
 * @dev:
 */
static int
udc_proc_read(char *page, char **start, off_t off, int count,
	      int *eof, void *_dev)
{
	char *buf = page;
	char *next = buf;
	unsigned size = count;
	unsigned long flags;
	int t;

	DEBUG_ENTRY("%s\n", __FUNCTION__);

	if (off != 0)
		return 0;

	local_irq_save(flags);

	/* basic device status */
	t = scnprintf(next, size,
		      DRIVER_DESC "\n"
		      "%s version: %s\n\n",
		      driver_name, DRIVER_VERSION);
	size -= t;
	next += t;

	/* display usbcore registers */
	t = show_usbcore_register(next, size,
		"Root Control/Status      [HUB00] : %8.8x\n",
		ROOT_CONTROL);
	size -= t;
	next += t;

	t = show_usbcore_register(next, size,
		"Root Config Setting      [HUB04] : %8.8x\n",
		ROOT_CONF);
	size -= t;
	next += t;

	t = show_usbcore_register(next, size,
		"Interrupt Control        [HUB08] : %8.8x\n",
		INT_CONTROL);
	size -= t;
	next += t;

	t = show_usbcore_register(next, size,
		"Interrupt Status         [HUB0c] : %8.8x\n",
		INT_STATUS);
	size -= t;
	next += t;

	t = show_usbcore_register(next, size,
		"Endpoint ACK Intr Enable [HUB10] : %8.8x\n",
		EP_ACK_INT_CONTROL);
	size -= t;
	next += t;

	t = show_usbcore_register(next, size,
		"Endpoint NAK Intr Enable [HUB14] : %8.8x\n",
		EP_NAK_INT_CONTROL);
	size -= t;
	next += t;

	t = show_usbcore_register(next, size,
		"Endpoint ACK Intr Status [HUB18] : %8.8x\n",
		EP_ACK_INT_STATUS);
	size -= t;
	next += t;

	t = show_usbcore_register(next, size,
		"Endpoint NAK Intr Status [HUB1c] : %8.8x\n",
		EP_NAK_INT_STATUS);
	size -= t;
	next += t;

	t = show_usbcore_register(next, size,
		"Device Soft Reset Enable [HUB20] : %8.8x\n",
		SW_RESET);
	size -= t;
	next += t;

	t = show_usbcore_register(next, size,
		"USB Status               [HUB24] : %8.8x\n",
		USB_STATUS);
	size -= t;
	next += t;

	t = show_usbcore_register(next, size,
		"Endpoint Data Toggle     [HUB28] : %8.8x\n",
		EP_TOGGLE);
	size -= t;
	next += t;

	t = show_usbcore_register(next, size,
		"Isochronous Transaction  [HUB2c] : %8.8x\n",
		ISO_FAIL_ACC);
	size -= t;
	next += t;

	t = show_usbcore_register(next, size,
		"EP 0 Control/Status      [HUB30] : %8.8x\n",
		EP0_CONTROL);
	size -= t;
	next += t;

	t = show_usbcore_register(next, size,
		"EP 0 Buffer Address      [HUB34] : %8.8x\n",
		EP0_BASE);
	size -= t;
	next += t;

	t = show_usbcore_register(next, size,
		"EP 1 Control/Status      [HUB38] : %8.8x\n",
		EP1_CONTROL);
	size -= t;
	next += t;

	t = show_usbcore_register(next, size,
		"EP 1 Status Bitmap       [HUB3c] : %8.8x\n",
		EP1_BITMAP);
	size -= t;
	next += t;

	t = show_usbcore_register_array(next, size,
		"Device Function Enable   [DEV00] : %8.8x %8.8x %8.8x %8.8x\n",
		DEVICE_CONTROL, USBCORE_DEVICE_SIZE);
	size -= t;
	next += t;

	t = show_usbcore_register_array(next, size,
		"                                 : %8.8x %8.8x %8.8x %8.8x\n",
		DEVICE_CONTROL + (4*USBCORE_DEVICE_SIZE),
		USBCORE_DEVICE_SIZE);
	size -= t;
	next += t;

	t = show_usbcore_register_array(next, size,
		"Device Interrupt Status  [DEV04] : %8.8x %8.8x %8.8x %8.8x\n",
		DEVICE_INT_STATUS, USBCORE_DEVICE_SIZE);
	size -= t;
	next += t;

	t = show_usbcore_register_array(next, size,
		"                                 : %8.8x %8.8x %8.8x %8.8x\n",
		DEVICE_INT_STATUS + (4*USBCORE_DEVICE_SIZE),
		USBCORE_DEVICE_SIZE);
	size -= t;
	next += t;

	t = show_usbcore_register_array(next, size,
		"Device Control Status    [DEV08] : %8.8x %8.8x %8.8x %8.8x\n",
		DEVICE_EP0_CTRL, USBCORE_DEVICE_SIZE);
	size -= t;
	next += t;

	t = show_usbcore_register_array(next, size,
		"                                 : %8.8x %8.8x %8.8x %8.8x\n",
		DEVICE_EP0_CTRL + (4*USBCORE_DEVICE_SIZE),
		USBCORE_DEVICE_SIZE);
	size -= t;
	next += t;

	t = show_usbcore_register_array(next, size,
		"Endpoint Configuration   [EPP00] : %8.8x %8.8x %8.8x %8.8x\n",
		EP_CONTROL, USBCORE_ENDPOINT_SIZE);
	size -= t;
	next += t;

	t = show_usbcore_register_array(next, size,
		"                                 : %8.8x %8.8x %8.8x %8.8x\n",
		EP_CONTROL + (4*USBCORE_ENDPOINT_SIZE),
		USBCORE_ENDPOINT_SIZE);
	size -= t;
	next += t;

	t = show_usbcore_register_array(next, size,
		"Endpoint Descriptor      [EPP04] : %8.8x %8.8x %8.8x %8.8x\n",
		EP_LIST_CONTROL, USBCORE_ENDPOINT_SIZE);
	size -= t;
	next += t;

	t = show_usbcore_register_array(next, size,
		"                                 : %8.8x %8.8x %8.8x %8.8x\n",
		EP_LIST_CONTROL + (4*USBCORE_ENDPOINT_SIZE),
		USBCORE_ENDPOINT_SIZE);
	size -= t;
	next += t;

	t = show_usbcore_register_array(next, size,
		"Endpoint Descriptor Base [EPP08] : %8.8x %8.8x %8.8x %8.8x\n",
		EP_BASE, USBCORE_ENDPOINT_SIZE);
	size -= t;
	next += t;

	t = show_usbcore_register_array(next, size,
		"                                 : %8.8x %8.8x %8.8x %8.8x\n",
		EP_BASE + (4*USBCORE_ENDPOINT_SIZE),
		USBCORE_ENDPOINT_SIZE);
	size -= t;
	next += t;

	t = show_usbcore_register_array(next, size,
		"Endpoint Descriptor List [EPP0c] : %8.8x %8.8x %8.8x %8.8x\n",
		EP_LIST_PTRS, USBCORE_ENDPOINT_SIZE);
	size -= t;
	next += t;

	t = show_usbcore_register_array(next, size,
		"                                 : %8.8x %8.8x %8.8x %8.8x\n",
		EP_LIST_PTRS + (4*USBCORE_ENDPOINT_SIZE),
		USBCORE_ENDPOINT_SIZE);
	size -= t;
	next += t;

	local_irq_restore(flags);
	*eof = 1;
	return count - size;
}

#define create_proc_files() 	create_proc_read_entry(proc_node_name, 0, NULL, udc_proc_read, dev)
#define remove_proc_files() 	remove_proc_entry(proc_node_name, NULL)

#else	/* !CONFIG_USB_GADGET_DEBUG_FILES */

#define create_proc_files() do {} while (0)
#define remove_proc_files() do {} while (0)

#endif	/* CONFIG_USB_GADGET_DEBUG_FILES */

/*
 * gadget api routines
 */

/**
 * usb_gadget_register_driver - register entry point for peripheral controller
 * @driver:
 */
int usb_gadget_register_driver(struct usb_gadget_driver *driver)
{
	struct aspeed_udc *dev = the_controller;
	int retval;

	DEBUG_ENTRY("%s: %s\n", __FUNCTION__, driver->driver.name);

	if (!driver
	    || !driver->bind
	    || !driver->unbind
	    || !driver->setup) {
		printk("usb_udc: register_driver return EINVAL\n");
		return -EINVAL;
		}
	if (!dev) {
		printk("usb_udc: register_driver return ENODEV\n");
		return -ENODEV;
	}
	if (dev->driver) {
		printk("usb_udc: register_driver return EBUSY\n");
		return -EBUSY;
	}

	/* first hook up the driver ... */
	driver->driver.bus = NULL;
	dev->driver = driver;
	dev->gadget.dev.driver = &driver->driver;

	retval = device_add(&dev->gadget.dev);
	retval = driver->bind(&dev->gadget);
	if (retval) {
		printk("usb_udc: %s: bind to driver %s --> error %d\n", dev->gadget.name,
		       driver->driver.name, retval);
		device_del(&dev->gadget.dev);

		dev->driver = 0;
		dev->gadget.dev.driver = 0;
		return retval;
	}

	/* ... then enable host detection and ep0; and we're ready
	 * for set_configuration as well as eventual disconnect.
	 * NOTE:  this shouldn't power up until later.
	 */
	printk("usb_udc: %s: registered gadget driver '%s'\n", dev->gadget.name,
	       driver->driver.name);

	udc_enable(dev);

	return 0;
}

EXPORT_SYMBOL(usb_gadget_register_driver);

/**
 * usb_gadget_unregister_driver - unregister entry point 
 * @driver:
 */
int usb_gadget_unregister_driver(struct usb_gadget_driver *driver)
{
	struct aspeed_udc *dev = the_controller;
	unsigned long flags;

	if (!dev)
		return -ENODEV;
	if (!driver || driver != dev->driver)
		return -EINVAL;

	spin_lock_irqsave(&dev->lock, flags);
	dev->driver = 0;
	udc_stop_activity(dev, driver);
	spin_unlock_irqrestore(&dev->lock, flags);

	driver->unbind(&dev->gadget);
	device_del(&dev->gadget.dev);

	udc_disable(dev);

	DEBUG_PRINT("unregistered gadget driver '%s'\n", driver->driver.name);
	return 0;
}

EXPORT_SYMBOL(usb_gadget_unregister_driver);

/** 
 * aspeed_ep_enable - enable endpoint
 * @_ep: endpoint
 * @desc: descriptor
 *
 * endpoint is assigned to unique device controller
 */
static int aspeed_ep_enable(struct usb_ep *_ep,
			     const struct usb_endpoint_descriptor *desc)
{
	struct aspeed_ep *ep;
	struct aspeed_udc *dev;
	unsigned long flags;
	int index, ret;

	DEBUG_ENTRY("%s, %p\n", __FUNCTION__, _ep);

	ep = container_of(_ep, struct aspeed_ep, ep);
	if (!_ep || !desc || ep->desc || _ep->name == ep0name
           || desc->bDescriptorType != USB_DT_ENDPOINT
            || ep->endpointAddress != desc->bEndpointAddress) {
		printk("usb_udc: %s, bad ep or descriptor\n", __FUNCTION__);
		return -EINVAL;
	}

	index = ep_index(ep);
	if ((index == 0) || (index >= UDC_MAX_ENDPOINTS)) {
		printk("usb_udc: %s, bad endpoint index %d\n", __FUNCTION__, index);
		return -EINVAL;
	}

	/* xfer types must match, except that interrupt ~= bulk */
	if (ep->bmAttributes != desc->bmAttributes
	    && ep->bmAttributes != USB_ENDPOINT_XFER_BULK
	    && desc->bmAttributes != USB_ENDPOINT_XFER_INT) {
		printk("usb_udc: %s, %s type mismatch\n", __FUNCTION__, _ep->name);
		return -EINVAL;
	}

	/* set maximum packet size */
	if (ep->bmAttributes == USB_ENDPOINT_XFER_BULK)
		ep->ep.maxpacket = MAX_PACKET_SIZE_BULK;
	else if (ep->bmAttributes == USB_ENDPOINT_XFER_INT)
		ep->ep.maxpacket = MAX_PACKET_SIZE_INT;
	else	ep->ep.maxpacket = MAX_PACKET_SIZE_CONTROL;

	if (le16_to_cpu(desc->wMaxPacketSize) > ep_maxpacket(ep)) {
		printk("usb_udc: %s, bad %s maxpacket\n", __FUNCTION__, _ep->name);
		return -ERANGE;
	}

	dev = ep->dev;
	if (!dev->driver || dev->gadget.speed == USB_SPEED_UNKNOWN) {
		printk("usb_udc: %s, bogus device state\n", __FUNCTION__);
		return -ESHUTDOWN;
	}

	spin_lock_irqsave(&ep->dev->lock, flags);

	ep->stopped = 0;
	ep->desc = desc;

	ep->ep.maxpacket = le16_to_cpu(desc->wMaxPacketSize);

	ret = enable_endpoint(ep);
	if (ret) {
		printk("usb_udc: %s: enable failed %d\n", __FUNCTION__, ret);
		return ret;
	}
	ep->enabled++;

	/* Reset halt state (does flush) */
	aspeed_set_halt(_ep, 0);
	spin_unlock_irqrestore(&ep->dev->lock, flags);

	DEBUG_PRINT("%s: enabled %s\n", __FUNCTION__, _ep->name);
	return 0;
}

/** 
 * aspeed_ep_disable - disable endpoint
 * @_ep: endpoint
 */
static int aspeed_ep_disable(struct usb_ep *_ep)
{
	struct aspeed_ep *ep;
	unsigned long flags;
	int index;

	DEBUG_ENTRY("%s, %p\n", __FUNCTION__, _ep);

	ep = container_of(_ep, struct aspeed_ep, ep);
	if (!_ep || !ep->desc) {
		printk("usb_udc: %s, %s not enabled\n", __FUNCTION__,
		      _ep ? ep->ep.name : NULL);
		return -EINVAL;
	}

	index = ep_index(ep);
	if ((index == 0) || (index >= UDC_MAX_ENDPOINTS)) {
		printk("usb_udc: %s, bad endpoint index %d\n", __FUNCTION__, index);
		return -EINVAL;
	}

	spin_lock_irqsave(&ep->dev->lock, flags);

	/* Nuke all pending requests (does flush) */
	udc_nuke(ep, -ESHUTDOWN);

	/* Disable ep */
	ep->desc = 0;
	ep->stopped = 1;
	ep->enabled = 0;

	disable_endpoint(ep);

	spin_unlock_irqrestore(&ep->dev->lock, flags);

	DEBUG_PRINT("%s: disabled %s\n", __FUNCTION__, _ep->name);
	return 0;
}

/**
 * aspeed_alloc_request - allocate request structure
 * @ep: endpoint structure
 * @gfp_flags: allocation flags
 */
static struct usb_request *aspeed_alloc_request(struct usb_ep *ep,
						 gfp_t gfp_flags)
{
	struct aspeed_request *req;

	DEBUG_ENTRY("%s, %p\n", __FUNCTION__, ep);

	req = kzalloc(sizeof(*req), gfp_flags);
	if (!req)
		return 0;

	DEBUG_LIST("%s request allocated %p\n", __FUNCTION__, ep);
	INIT_LIST_HEAD(&req->queue);

	return &req->req;
}

/**
 * aspeed_free_request - release request structure
 * @ep: endpoint structure
 * @_req: request structure to free
 */
static void aspeed_free_request(struct usb_ep *ep, struct usb_request *_req)
{
	struct aspeed_request *req;

	DEBUG_ENTRY("%s, %p\n", __FUNCTION__, ep);

	req = container_of(_req, struct aspeed_request, req);
	WARN_ON(!list_empty(&req->queue));
	DEBUG_LIST("%s, free request %p\n", __FUNCTION__, req);
	kfree(req);
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,23)
/**
 * aspeed_alloc_buffer - allocate buffer
 * @ep: endpoint structure
 * @bytes: size 
 * @dma: dma address (on return)
 * @gfp_flags: allocation flags
 */
static void *aspeed_alloc_buffer(struct usb_ep *ep, unsigned bytes,
				  dma_addr_t * dma, gfp_t gfp_flags)
{
	void * buf;

	DEBUG_ENTRY("%s (%p, %d, %d)\n", __FUNCTION__, ep, bytes, gfp_flags);

	*dma = (dma_addr_t) NULL;
	buf = dma_alloc_coherent(the_controller->gadget.dev.parent,
		 bytes, dma, gfp_flags);

	DEBUG_LIST("%s, allocated %p (%p)\n", __FUNCTION__, buf, *dma);

	return buf;
}

/**
 * aspeed_free_buffer - release buffer
 * @ep: endpoint structure
 * @buf: buffer to release
 * @dma: dma address
 * @bytes: size 
 */
static void aspeed_free_buffer(struct usb_ep *ep, void *buf, dma_addr_t dma,
				unsigned bytes)
{
	DEBUG_ENTRY("%s, %p\n", __FUNCTION__, ep);
	DEBUG_LIST("%s, free %p\n", __FUNCTION__, buf);

	if (dma != (dma_addr_t) NULL) 
		dma_free_coherent(the_controller->gadget.dev.parent,
			bytes, buf, dma);
}
#endif

/**
 * aspeed_queue - queue single request
 * @_ep: endpoint structure
 * @req: request
 * @gfp: allocation flags
 */
static int aspeed_queue(struct usb_ep *_ep, struct usb_request *_req,
			 gfp_t gfp_flags)
{
	struct aspeed_request *req;
	struct aspeed_ep *ep;
	struct aspeed_udc *dev;
	unsigned long flags;

	DEBUG_ENTRY("%s, %p\n", __FUNCTION__, _ep);
	DEBUG_LIST("%s, queue %p\n", __FUNCTION__, _req);

	req = container_of(_req, struct aspeed_request, req);
	if (unlikely
	    (!_req || !_req->complete || !_req->buf
	     || !list_empty(&req->queue))) {
		printk("usb_udc: %s, bad params\n", __FUNCTION__);
		return -EINVAL;
	}

	ep = container_of(_ep, struct aspeed_ep, ep);
	if (unlikely(!_ep || (!ep->desc && ep->ep.name != ep0name))) {
		printk("usb_udc: %s, bad ep\n", __FUNCTION__);
		return -EINVAL;
	}

	dev = ep->dev;
	if (unlikely(!dev->driver || dev->gadget.speed == USB_SPEED_UNKNOWN)) {
		printk("usb_udc: %s, bogus device state %p\n", __FUNCTION__, dev->driver);
		return -ESHUTDOWN;
	}

	DEBUG_LIST("%s queue req %p, len %d buf %p\n", _ep->name, _req,
		_req->length, _req->buf);

	spin_lock_irqsave(&dev->lock, flags);

	_req->status = -EINPROGRESS;
	_req->actual = 0;

	/* kickstart this i/o queue? */
	DEBUG_LIST("Add to Q ep[%d] empty %d\n", ep_index(ep),
		list_empty(&ep->queue));

	if (list_empty(&ep->queue) && likely(!ep->stopped)) {
		if (unlikely(ep_index(ep) == 0)) {
			list_add_tail(&req->queue, &ep->queue);
			udc_ep0_kick(dev, ep);
		} else if (ep_is_in(ep)) {
			list_add_tail(&req->queue, &ep->queue);
			if (ep->transmitBusy == 0)
				write_fifo(ep, req, 0);
		} else {
			list_add_tail(&req->queue, &ep->queue);
			ep->xfer_ops->ready_to_receive(ep);
//			read_fifo(ep, req);
		}
	} else {
		list_add_tail(&req->queue, &ep->queue);
	}

	spin_unlock_irqrestore(&dev->lock, flags);
	return 0;
}

/**
 * aspeed_dequeue - dequeue single request
 * @_ep: endpoint structure
 * @req: request
 */
static int aspeed_dequeue(struct usb_ep *_ep, struct usb_request *_req)
{
	struct aspeed_ep *ep;
	struct aspeed_request *req;
	unsigned long flags;

	DEBUG_PRINT("%s, %p\n", __FUNCTION__, _ep);
	DEBUG_LIST("%s, dequeue %p\n", __FUNCTION__, _req);

	ep = container_of(_ep, struct aspeed_ep, ep);
	if (!_ep || ep->ep.name == ep0name)
		return -EINVAL;

	spin_lock_irqsave(&ep->dev->lock, flags);

	/* make sure it's actually queued on this endpoint */
	list_for_each_entry(req, &ep->queue, queue) {
		if (&req->req == _req)
			break;
	}
	if (&req->req != _req) {
		spin_unlock_irqrestore(&ep->dev->lock, flags);
		return -EINVAL;
	}

	udc_done(ep, req, -ECONNRESET);

	spin_unlock_irqrestore(&ep->dev->lock, flags);
	return 0;
}

static void reset_data_toggle(struct aspeed_ep * ep);

/**
 * aspeed_set_halt - halt specified endpoint
 * @ep: endpoint
 * @value: 1 to set halt, 0 to clear halt
 *
 * Return 0 if success
 */
static int aspeed_set_halt(struct usb_ep *_ep, int value)
{
	struct aspeed_ep *ep;
	unsigned long flags;
	int ret = 0;

	ep = container_of(_ep, struct aspeed_ep, ep);

	DEBUG_ENTRY("%s, ep %d, val %d\n", __FUNCTION__, ep_index(ep),
		value);

	spin_lock_irqsave(&ep->dev->lock, flags);

#ifdef	STALL_SUPPORT
	if (value)
		ret = set_stall(ep);
	else	ret = clear_stall(ep);

	if (ret == 0) {
		if (value) {
			ep->stopped = 1;
		} else {
			ep->stopped = 0;
		}
	}
#else

	if (!value)
		reset_data_toggle(ep);
#endif

	spin_unlock_irqrestore(&ep->dev->lock, flags);

	DEBUG_PRINT("%s %s halted\n", _ep->name, value == 0 ? "NOT" : "IS");
	return ret;
}

/**
 * aspeed_fifo_status - return bytes in fifo
 * @_ep: endpoint
 *
 * Returns count
 */
static int aspeed_fifo_status(struct usb_ep *_ep)
{
	DEBUG_PRINT("%s\n", __FUNCTION__);

	/* currently not supported */
	return -EOPNOTSUPP;
}

/**
 * aspeed_fifo_flush - flush ep
 * @_ep: endpoint
 * Returns count
 */
static void aspeed_fifo_flush(struct usb_ep *_ep)
{
	struct aspeed_ep *ep;

	DEBUG_ENTRY("%s\n", __FUNCTION__);

	ep = container_of(_ep, struct aspeed_ep, ep);
	if (unlikely(!_ep || (!ep->desc && ep->ep.name != ep0name))) {
		DEBUG_PRINT("%s, bad ep\n", __FUNCTION__);
		return;
	}

	udc_flush(ep);
}

static void nop_release(struct device *dev)
{
	DEBUG_ENTRY("%s %s\n", __FUNCTION__, dev->bus_id);
}

/*
 * device-scoped parts of the api to the usb controller hardware
 */

static int aspeed_udc_get_frame(struct usb_gadget *_gadget)
{
	DEBUG_ENTRY("%s, %p\n", __FUNCTION__, _gadget);
	return 0;

}

static int aspeed_udc_wakeup(struct usb_gadget *_gadget)
{
	DEBUG_ENTRY("%s, %p\n", __FUNCTION__, _gadget);
	return 0;
}

static int aspeed_udc_set_selfpowered(struct usb_gadget *_gadget, int value)
{
	DEBUG_ENTRY("%s, %p\n", __FUNCTION__, _gadget);
	return 0;
}

/**
 * udc_pullup - device disconnect / connect
 * @_gadget: usb slave structure
 * @is_on: true when device connected
 */
static int aspeed_udc_pullup(struct usb_gadget *_gadget, int is_on)
{
	unsigned long flags;
	int i;
	struct aspeed_ep *ep;

	DEBUG_ENTRY("%s, %p\n", __FUNCTION__, _gadget);

	spin_lock_irqsave(&the_controller->lock, flags);

	if (is_on) {
		if (the_controller->core_initialized == 0)
			usbcore_start();
		for (i=0; i<UDC_MAX_ENDPOINTS; i++) {
			ep = &the_controller->ep[i];
			if (ep->enabled)
				enable_endpoint(ep);
		}
		the_controller->gadget.speed = USB_SPEED_UNKNOWN;
	} else {
		for (i=0; i<UDC_MAX_ENDPOINTS; i++) {
			ep = &the_controller->ep[i];
			udc_nuke(ep, -ESHUTDOWN);
			disable_endpoint(ep);
		}
		the_controller->gadget.speed = USB_SPEED_UNKNOWN;
		usbcore_stop();
	}
	
	spin_unlock_irqrestore(&the_controller->lock, flags);
	
	if (!is_on && the_controller->driver != NULL)
		the_controller->driver->disconnect(_gadget);

	return 0;
}

static const struct usb_gadget_ops aspeed_udc_ops = {
	.get_frame = aspeed_udc_get_frame,
	.wakeup = aspeed_udc_wakeup,
	.set_selfpowered = aspeed_udc_set_selfpowered,
	.pullup = aspeed_udc_pullup,
};

/**
 * udc_controller - endpoint definitions
 *
 * initializer for endpoints
 */
static struct aspeed_udc udc_controller = {
	.gadget = {
		   .ops = &aspeed_udc_ops,
		   .ep0 = &udc_controller.ep[0].ep,
		   .name = driver_name,
		   .dev = {
			   .bus_id = "gadget",
			   .release = nop_release,
			   },
		   },

	/* control endpoint */
	.ep[0] = {
		  .ep = {
			 .name = ep0name,
			 .ops = &aspeed_ep_ops,
			 .maxpacket = 64,
			 },
		  .dev = &udc_controller,
		  .xfer_ops = &transfer_ops_control,
		  .ep_type = ep_control,
		  .endpointAddress = 0,
		  .bmAttributes = 0,
		  },

	/* endpoint 1 - in interrupt */
	.ep[1] = {
		  .ep = {
			 .name = "ep1in-int",
			 .ops = &aspeed_ep_ops,
			 .maxpacket = MAX_PACKET_SIZE_INT,
			 },
		  .dev = &udc_controller,
		  .xfer_ops = &transfer_ops_int,
		  .ep_type = ep_interrupt,
		  .endpointAddress = 1 | USB_DIR_IN,
		  .bmAttributes = USB_ENDPOINT_XFER_INT,
		  },

	/* endpoint 2 - in interrupt */
	.ep[2] = {
		  .ep = {
			 .name = "ep2in-int",
			 .ops = &aspeed_ep_ops,
			 .maxpacket = MAX_PACKET_SIZE_INT,
			 },
		  .dev = &udc_controller,
		  .xfer_ops = &transfer_ops_int,
		  .ep_type = ep_interrupt,
		  .endpointAddress = 2 | USB_DIR_IN,
		  .bmAttributes = USB_ENDPOINT_XFER_INT,
		  },

	/* endpoint 3 - in interrupt */
	.ep[3] = {
		  .ep = {
			 .name = "ep3in-int",
			 .ops = &aspeed_ep_ops,
			 .maxpacket = MAX_PACKET_SIZE_INT,
			 },
		  .dev = &udc_controller,
		  .xfer_ops = &transfer_ops_int,
		  .ep_type = ep_interrupt,
		  .endpointAddress = 3 | USB_DIR_IN,
		  .bmAttributes = USB_ENDPOINT_XFER_INT,
		  },

	/* endpoint 4 - in bulk-transfer */
	.ep[4] = {
		  .ep = {
			 .name = "ep4in-bulk",
			 .ops = &aspeed_ep_ops,
			 .maxpacket = MAX_PACKET_SIZE_BULK,
			 },
		  .dev = &udc_controller,
		  .xfer_ops = &transfer_ops_bulk,
		  .ep_type = ep_bulk_in,
		  .endpointAddress = 4 | USB_DIR_IN,
		  .bmAttributes = USB_ENDPOINT_XFER_BULK,
		  },

	/* endpoint 5 - in bulk-transfer */
	.ep[5] = {
		  .ep = {
			 .name = "ep5in-bulk",
			 .ops = &aspeed_ep_ops,
			 .maxpacket = MAX_PACKET_SIZE_BULK,
			 },
		  .dev = &udc_controller,
		  .xfer_ops = &transfer_ops_bulk,
		  .ep_type = ep_bulk_in,
		  .endpointAddress = 5 | USB_DIR_IN,
		  .bmAttributes = USB_ENDPOINT_XFER_BULK,
		  },

	/* endpoint 6 - out bulk-transfer */
	.ep[6] = {
		  .ep = {
			 .name = "ep6out-bulk",
			 .ops = &aspeed_ep_ops,
			 .maxpacket = MAX_PACKET_SIZE_BULK,
			 },
		  .dev = &udc_controller,
		  .xfer_ops = &transfer_ops_bulk,
		  .ep_type = ep_bulk_out,
		  .endpointAddress = 6,
		  .bmAttributes = USB_ENDPOINT_XFER_BULK,
		  },

	/* endpoint 7 - out bulk-transfer */
	.ep[7] = {
		  .ep = {
			 .name = "ep7out-bulk",
			 .ops = &aspeed_ep_ops,
			 .maxpacket = MAX_PACKET_SIZE_BULK,
			 },
		  .dev = &udc_controller,
		  .xfer_ops = &transfer_ops_bulk,
		  .ep_type = ep_bulk_out,
		  .endpointAddress = 7,
		  .bmAttributes = USB_ENDPOINT_XFER_BULK,
		  },
};

/**
 * aspeed_probe - device attach routine
 * @dev: device structure
 *
 * enable the device
 */
static int aspeed_probe(struct device *dev)
{
	struct aspeed_udc *ctrl = &udc_controller;
	struct platform_device *pdev;

        DEBUG_ENTRY("%s: %p\n", __FUNCTION__, dev);

	pdev = container_of(dev, struct platform_device, dev);

	the_controller = ctrl;

        dev_set_drvdata(dev, ctrl);
	ctrl->gadget.is_dualspeed = 1;

	device_initialize(&ctrl->gadget.dev);
	strcpy(ctrl->gadget.dev.bus_id, "gadget");

	udc_reinit(ctrl);
	udc_enable(ctrl);
    	create_proc_files();

	return 0;
}

/**
 * aspeed_remove - device removal routine
 * @dev: device structure
 *
 * disable the device
 */
static int aspeed_remove(struct device *dev)
{
	struct platform_device *pdev;

	DEBUG_ENTRY("%s: %p\n", __FUNCTION__, dev);

	pdev = container_of(dev, struct platform_device, dev);

	udc_disable(the_controller);
	remove_proc_files();
	the_controller = NULL;

	platform_set_drvdata(pdev, 0);

	return 0;
}

static void aspeed_release_platform(struct device  *dev)
{
}

static struct platform_device aspeed_usb_gadget_device = {
        .name           = (char *) driver_name,
        .id             = 0,
        .dev = {
                .dma_mask          = NULL,              /* not use dma */
                .coherent_dma_mask = 0xffffffff,
                .release           = aspeed_release_platform,
        },
};

static struct device_driver udc_driver = {
        .name   = (char *) driver_name,
        .bus    = &platform_bus_type,
        .probe  = aspeed_probe,
        .remove = aspeed_remove,
};

/*
 * standard udc routines
 */

/**
 * udc_disable - disable USB device controller
 * @dev: controller structure
 */
static void udc_disable(struct aspeed_udc *dev)
{
	DEBUG_ENTRY("%s, %p\n", __FUNCTION__, dev);

	if (udc_controller.core_initialized)
		disable_interrupts();

	dev->ep0state = WAIT_FOR_SETUP;
	dev->gadget.speed = USB_SPEED_UNKNOWN;
}

/**
 * udc_enable - enable USB device controller
 * @dev: controller structure
 */
static void udc_enable(struct aspeed_udc *dev)
{
	DEBUG_ENTRY("%s, %p\n", __FUNCTION__, dev);

	if (udc_controller.core_initialized)
		enable_interrupts();

	dev->ep0state = WAIT_FOR_SETUP;
}

/**
 * udc_reinit - initialize software state
 * @dev: controller structure
 */
static void udc_reinit(struct aspeed_udc *dev)
{
	u32 i;

	DEBUG_ENTRY("%s, %p\n", __FUNCTION__, dev);

	/* device/ep0 records init */
	DEBUG_LIST("%s, initialize\n", __FUNCTION__);
	INIT_LIST_HEAD(&dev->gadget.ep_list);
	INIT_LIST_HEAD(&dev->gadget.ep0->ep_list);
	dev->ep0state = WAIT_FOR_SETUP;

	/* basic endpoint records init */
	for (i = 0; i < UDC_MAX_ENDPOINTS; i++) {
		struct aspeed_ep *ep = &dev->ep[i];

		if (i != 0)
			list_add_tail(&ep->ep.ep_list, &dev->gadget.ep_list);

		ep->desc = 0;
		INIT_LIST_HEAD(&ep->queue);
	}
}

/**
 * udc_ep0_kick - kick endpoint 0 to do next transfer
 * @dev: controller structure
 * @ep: endpoint structure
 */
static void udc_ep0_kick(struct aspeed_udc *dev, struct aspeed_ep *ep)
{
	DEBUG_ENTRY("%s, %p\n", __FUNCTION__, ep);

	/* Clear "out packet ready" */
	ep->xfer_ops->ready_to_receive(ep);

	if (ep_is_in(ep)) {
		dev->ep0state = DATA_STATE_CONTROL_READ;
		udc_ep0_in(dev, ep);
	} else {
		dev->ep0state = DATA_STATE_CONTROL_WRITE;
		udc_ep0_out(dev, ep);
	}
}

/**
 * udc_ep0_in - transmit routine for endpoint 0
 * @dev: controller structure
 * @ep: endpoint structure
 */
static int udc_ep0_in(struct aspeed_udc *dev, struct aspeed_ep *ep)
{
	struct aspeed_request *req;
	int ret, need_zlp = 0; 

	DEBUG_ENTRY("%s, %p\n", __FUNCTION__, ep);

	if (list_empty(&ep->queue))
		req = 0;
	else
		req = list_entry(ep->queue.next, struct aspeed_request, queue);

	if (!req) {
		DEBUG_EP0("%s: NULL REQ\n", __FUNCTION__);
		return 0;
	}

	if (req->req.length == 0) {
		ep->xfer_ops->transmit_in_packet(ep, req->req.buf, req->req.length);
		req->req.actual = 0;
		udc_done(ep, req, 0);
		DEBUG_EP0("%s: wait for setup\n", __FUNCTION__);
		dev->ep0state = WAIT_FOR_SETUP;
		return 1;
	}

	if (req->req.length - req->req.actual == ep_maxpacket(ep)) {
		/* next write will end with packet size */
		/* need zero-length packet */
		need_zlp = 1;
	}

	ret = write_fifo_ep0(ep, req);

	if (ret == 1 && !need_zlp) {
		/* last packet */
		DEBUG_EP0("%s: finished\n", __FUNCTION__);
		dev->ep0state = WAIT_FOR_SETUP;
	} else {
		DEBUG_EP0("%s: not finished\n", __FUNCTION__);
	}

	if (need_zlp) {
		DEBUG_EP0("%s: need zlp\n", __FUNCTION__);
		dev->ep0state = DATA_STATE_CONTROL_READ_ZLP;
	}

	return 1;
}

/**
 * udc_ep0_out - endpoint 0 receive routine
 * @dev: device structure
 * @ep: endpoint structure
 */
static void udc_ep0_out(struct aspeed_udc *dev, struct aspeed_ep *ep)
{
	struct aspeed_request *req;
	int ret;

	DEBUG_ENTRY("%s: %p\n", __FUNCTION__, ep);

	if (list_empty(&ep->queue))
		req = 0;
	else
		req = list_entry(ep->queue.next, struct aspeed_request, queue);

	if (req) {

		if (req->req.length == 0) {
			DEBUG_EP0("ZERO LENGTH OUT!\n");
			ep->xfer_ops->transmit_in_packet(ep, req->req.buf, req->req.length);
			req->req.actual = 0;
			udc_done(ep, req, 0);
			dev->ep0state = WAIT_FOR_SETUP;
			return;
		}

		ret = read_fifo_ep0(ep, req);
		if (ret) {
			/* Done! */
			DEBUG_EP0("%s: finished, waiting for status\n",
				  __FUNCTION__);
			dev->ep0state = WAIT_FOR_CONTROL_WRITE_STATUS;
		} else {
			/* Not done yet.. */
			DEBUG_EP0("%s: not finished\n", __FUNCTION__);
		}
	} else {
		DEBUG_EP0("NO REQ??!\n");
		ret = ep->xfer_ops->recv_out_packet(ep, udc_controller.control_buf, 64);
		udc_controller.control_buf_count = ret;
	}
}

/**
 * udc_done - retire a request; caller blocked irqs
 * @ep: endpoint structure
 * @req: request structure
 * @status: status to return
 */
static void udc_done(struct aspeed_ep *ep, struct aspeed_request *req,
	int status)
{
	unsigned int stopped = ep->stopped;

	DEBUG_ENTRY("%s, %p\n", __FUNCTION__, ep);
	DEBUG_LIST("%s, delete %p\n", __FUNCTION__, req);
	list_del_init(&req->queue);

	if (likely(req->req.status == -EINPROGRESS))
		req->req.status = status;
	else
		status = req->req.status;

	if (status && status != -ESHUTDOWN)
		DEBUG_PRINT("complete %s req %p stat %d len %u/%u\n",
		      ep->ep.name, &req->req, status,
		      req->req.actual, req->req.length);

	/* don't modify queue heads during completion callback */
	ep->stopped = 1;

	spin_unlock(&ep->dev->lock);
	req->req.complete(&ep->ep, &req->req);
	spin_lock(&ep->dev->lock);

	ep->stopped = stopped;
}

/**
 * udc_nuke - dequeue all requests
 * @ep: endpoint structure
 * @status: status to return
 */
void udc_nuke(struct aspeed_ep *ep, int status)
{
	struct aspeed_request *req;

	DEBUG_ENTRY("%s, %p\n", __FUNCTION__, ep);

	/* Flush FIFO */
	udc_flush(ep);

	/* called with irqs blocked */
	while (!list_empty(&ep->queue)) {
		req = list_entry(ep->queue.next, struct aspeed_request, queue);
		udc_done(ep, req, status);
	}

}

/**
 * udc_flush - flush all queued requests
 * @ep: endpoint structure
 */
static void udc_flush(struct aspeed_ep *ep)
{
	DEBUG_ENTRY("%s, %p\n", __FUNCTION__, ep);

	cancel_delayed_work(&ep->work);
	if (udc_controller.core_initialized)
		ep->xfer_ops->terminate_in_packet(ep);
}

/**
 * udc_stop_activity - stop all driver activity
 * @dev:
 * @driver:
 */
static void udc_stop_activity(struct aspeed_udc *dev,
			  struct usb_gadget_driver *driver)
{
	int i;

	DEBUG_ENTRY("%s\n", __FUNCTION__);

	/* don't disconnect drivers more than once */
	if (dev->gadget.speed == USB_SPEED_UNKNOWN)
		driver = 0;
	dev->gadget.speed = USB_SPEED_UNKNOWN;

	/* prevent new request submissions, kill any outstanding requests  */
	for (i = 0; i < UDC_MAX_ENDPOINTS; i++) {
		struct aspeed_ep *ep = &dev->ep[i];
		ep->stopped = 1;
		udc_nuke(ep, -ESHUTDOWN);
	}

	/* report disconnect; the driver is already quiesced */
	if (driver) {
		spin_unlock(&dev->lock);
		driver->disconnect(&dev->gadget);
		spin_lock(&dev->lock);
	}

	/* re-init driver-visible data structures */
	udc_reinit(dev);
}

/*
 * fifo routines
 */

/**
 * write_fifo - send packet from current request
 * @ep: endpoint structure
 * @req: request structure
 * @busy: busy state
 *
 * returns 0 = still running, 1 = complete, negative = errno
 */
static int write_fifo(struct aspeed_ep *ep, struct aspeed_request *req,
	int busy)
{
	int max;
	int is_last, length;
	u8 *buf;

	DEBUG_ENTRY("%s, %p\n", __FUNCTION__, ep);

	max = le16_to_cpu(ep->desc->wMaxPacketSize);

	if ((busy == 0) && ep->transmitBusy) {
		printk("usb_udc: unexpected transmit busy in write_fifo\n");
	}	

	if (busy && (ep->transmitBusy == 0)) {
		printk("usb_udc: unexpected transmit NOT busy in write_fifo\n");
	}
	
	ep->transmitBusy++;

	buf = req->req.buf + req->req.actual;
	length = req->req.length - req->req.actual;
	length = min(length, max);

	ep->xfer_ops->transmit_in_packet(ep, buf, length);

	req->req.actual += length;

	/* last packet is usually short (or a zlp) */
	if (unlikely(length != max))
		is_last = 1;
	else {
		if (req->req.length != req->req.actual)
			is_last = 0;
		else	is_last = 1;
	}

	DEBUG_PRINT("%s: wrote %s %d bytes%s %d left\n",
		__FUNCTION__, ep->ep.name, length, is_last ? "/L" : "",
		req->req.length - req->req.actual);

	if (is_last)  {
		udc_done(ep, req, 0);
		return 1;
	} else	return 0;
}

/**
 * write_fifo_ep0 - send packet from current request, endpoint 0
 * @ep: endpoint structure
 * @req: request structure
 *
 * returns 0 = still running, 1 = complete, negative = errno
 */
static int write_fifo_ep0(struct aspeed_ep *ep, struct aspeed_request *req)
{
	int max;
	int is_last, length;
	u8 *buf;

	DEBUG_ENTRY("%s, %p\n", __FUNCTION__, ep);

	max = ep_maxpacket(ep);

	buf = req->req.buf + req->req.actual;
	length = req->req.length - req->req.actual;
	length = min(length, max);

	ep->xfer_ops->transmit_in_packet(ep, buf, length);

	req->req.actual += length;

	/* last packet is usually short (or a zlp) */
	if (unlikely(length != max))
		is_last = 1;
	else {
		if ((req->req.length != req->req.actual) || req->req.zero)
			is_last = 0;
		else	is_last = 1;
	}

	DEBUG_EP0("%s: wrote %s %d bytes%s %d left\n",
		__FUNCTION__, ep->ep.name, length, is_last ? "/L" : "",
		req->req.length - req->req.actual);

	if (is_last) {
		udc_done(ep, req, 0);
		return 1;
	}
	else	return 0;
}

/**
 * read_fifo - recv packet to current request
 * @ep: endpoint structure
 * @req: request structure
 *
 * returns 0 = still running, 1 = complete, negative = errno
 */
static int read_fifo(struct aspeed_ep *ep, struct aspeed_request *req)
{
	int actual, is_short;
	u8 *buf;
	u32 count, bufferspace;

	DEBUG_ENTRY("%s, %p\n", __FUNCTION__, ep);

	count = ep->xfer_ops->data_in_fifo(ep);
	if (count == 0) {
		printk("usb_udc: %s: Packet NOT ready\n", __FUNCTION__);
		return -EINVAL;
	}

	buf = req->req.buf + req->req.actual;
	bufferspace = req->req.length - req->req.actual;
	count = min(count, bufferspace);

	actual = ep->xfer_ops->recv_out_packet(ep, buf, count);
	req->req.actual += actual;
	is_short = (actual < ep->ep.maxpacket);

	DEBUG_PRINT("%s: read %s %d bytes%s req length %d\n",
		__FUNCTION__, ep->ep.name, actual,
		is_short ? "/S" : "", req->req.length);

	if (is_short || req->req.actual == req->req.length) {
		udc_done(ep, req, 0);
		return 1;
	} else {
		ep->xfer_ops->ready_to_receive(ep);
		return 0;
	}
}

/**
 * read_fifo_ep0 - recv packet to current request, endpoint 0
 * @ep: endpoint structure
 * @req: request structure
 *
 * returns 0 = still running, 1 = complete, negative = errno
 */
static int read_fifo_ep0(struct aspeed_ep *ep, struct aspeed_request *req)
{
	int actual;
	u8 *buf;
	u32 count, bufferspace;

	DEBUG_ENTRY("%s, %p\n", __FUNCTION__, ep);

	count = ep->xfer_ops->data_in_fifo(ep);

	buf = req->req.buf + req->req.actual;
	bufferspace = req->req.length - req->req.actual;
	count = min(count, bufferspace);

	actual = ep->xfer_ops->recv_out_packet(ep, buf, count);
	req->req.actual += actual;

	DEBUG_PRINT("%s: read %s %d bytes req length %d\n",
		__FUNCTION__, ep->ep.name, actual, req->req.length);

	if (req->req.actual == req->req.length) {
		udc_done(ep, req, 0);
		return 1;
	} else	return 0;
}

/*
 * low-level aspeed 2050 usbcore routines
 */

/**
 * read_setup_packet - read endpoint EP0 control packet
 * @dev: controller structure
 * @ctrl: control request structure (returned data)
 */
static void read_setup_packet(struct aspeed_udc *dev,
		struct usb_ctrlrequest *ctrl)
{
	u32 recv[2];

	DEBUG_ENTRY("%s\n", __FUNCTION__);

	recv[0] = ioread32(dev->setup_low);
	recv[1] = ioread32(dev->setup_high);

	DEBUG_SETUP("setup packet: %x %x\n",
		recv[0], recv[1]);

	memcpy(ctrl, recv, sizeof(struct usb_ctrlrequest));
}

/**
 * disable_interrupts - disable all interrupts
 */
static void disable_interrupts(void)
{
	DEBUG_ENTRY("%s\n", __FUNCTION__);

	iowrite32(0,
		IO_ADDRESS(INT_CONTROL));
}

/**
 * enable_interrupts - enable all interrupts
 */
static void enable_interrupts(void)
{
	DEBUG_ENTRY("%s\n", __FUNCTION__);

	iowrite32(INT_CONTROL_INIT, IO_ADDRESS(INT_CONTROL));
	iowrite32(INT_STATUS_INIT, IO_ADDRESS(INT_STATUS));
}

/**
 * set_address - set the USB address for this device
 * @ep: endpoint structure
 * @address: address device number
 */
static void set_address(struct aspeed_ep * ep, u8 address)
{
	int index = ep_index(ep);

	DEBUG_ENTRY("%s, %p [ep %d]\n", __FUNCTION__, ep, index);

	if (index == 0) {
		/* EP 0 */
		iowrite32(address & ROOT_FUNCTION_DEVICE_ADDRESS,
			IO_ADDRESS(ROOT_CONF));	
	}
}

//#ifdef	STALL_SUPPORT	
/**
 * reset_data_toggle - reset data toggle for device to DATA0
 * @ep: endpoint structure
 */
static void reset_data_toggle(struct aspeed_ep * ep)
{
	int index = ep_index(ep);

	DEBUG_ENTRY("%s, %p [ep %d]\n", __FUNCTION__, ep, index);

	if (index) {
		/* EP 1 - 7 */
		iowrite32(index, IO_ADDRESS(EP_TOGGLE));
	}
}
//#endif

/**
 * enable_endpoint - enable specified endpoint
 * @ep: endpoint structure
 *
 * returns: 0 if ok, non-zero on failure
 */
static int enable_endpoint(struct aspeed_ep * ep)
{
	u32 cfg, intc;
	int index = ep_index(ep);

	DEBUG_ENTRY("%s, %p [ep %d]\n", __FUNCTION__, ep, index);

	if (index == 0)
		return 0;

	/* enable endpoint interrupts */
	intc = ioread32(IO_ADDRESS(EP_ACK_INT_CONTROL));
	intc |= (1 << index);
	iowrite32(intc, IO_ADDRESS(EP_ACK_INT_CONTROL));

	/* enable endpoint */
	iowrite32(0x4, IO_ADDRESS(EP_LIST_CONTROL +
			(index * USBCORE_ENDPOINT_SIZE)));
	iowrite32(0x2, IO_ADDRESS(EP_LIST_CONTROL +
			(index * USBCORE_ENDPOINT_SIZE)));
	udelay(1);

	iowrite32(ep->buffer_phy, IO_ADDRESS(EP_BASE +
			(index * USBCORE_ENDPOINT_SIZE)));

	iowrite32(0, IO_ADDRESS(EP_LIST_PTRS +
			(index * USBCORE_ENDPOINT_SIZE)));

	cfg = (EP_ENABLE);
	cfg |= (index << EP_NUMBER_SHFT);
	if (ep->ep_type == ep_interrupt)
		cfg |= EP_TYPE_INTERRUPT_IN;
	if (ep->ep_type == ep_bulk_in)
		cfg |= EP_TYPE_BULK_IN;
	if (ep->ep_type == ep_bulk_out)
		cfg |= EP_TYPE_BULK_OUT;
	iowrite32(cfg, IO_ADDRESS(EP_CONTROL +
		 (index * USBCORE_ENDPOINT_SIZE)));

	return 0;
}

/**
 * disable_endpoint - disable specified endpoint
 * @ep: endpoint structure
 */
static void disable_endpoint(struct aspeed_ep * ep)
{
	u32 cfg, intc;
	int index = ep_index(ep);

	DEBUG_ENTRY("%s, %p [ep %d]\n", __FUNCTION__, ep, index);

	if (index == 0)
		return;

	/* disable endpoint */
	cfg = ioread32(IO_ADDRESS(EP_CONTROL +
			(index * USBCORE_ENDPOINT_SIZE)));
	cfg &= ~EP_ENABLE;
	iowrite32(cfg, IO_ADDRESS(EP_CONTROL + 
			(index * USBCORE_ENDPOINT_SIZE)));

	/* disable endpoint interrupts */
	intc = ioread32(IO_ADDRESS(EP_ACK_INT_CONTROL));
	intc &= ~(1 << index);
	iowrite32(intc, IO_ADDRESS(EP_ACK_INT_CONTROL));
}

/**
 * get_speed - return USB speed
 */
static int get_speed(void)
{
	u32 status;

	status = ioread32(IO_ADDRESS(USB_STATUS));
	if (status & BUS_SPEED)
		return USB_SPEED_HIGH;
	else	return USB_SPEED_FULL;
}

#ifdef	STALL_SUPPORT	

/**
 * set_stall - stall specified endpoint
 * @ep: endpoint structure
 */
static int set_stall(struct aspeed_ep * ep)
{
	u32 ctrl, cfg;
	int index = ep_index(ep);

	DEBUG_ENTRY("%s, %p [ep %d]\n", __FUNCTION__, ep, index);

	if (index == 0) {
		/* EP0 */
		ctrl = ioread32(IO_ADDRESS(EP0_CONTROL));
		ctrl |= EP0_STALL_CONTROL;
		iowrite32(ctrl, IO_ADDRESS(EP0_CONTROL));
	} else {
		/* EP1 - EP7 */
		if (ep_is_in(ep)) {
			if ((!list_empty(&ep->queue)) ||
			    (ep->transmitBusy)) 
				return -EAGAIN;
		}

		cfg = ioread32(IO_ADDRESS(EP_CONTROL + 
				(index * USBCORE_ENDPOINT_SIZE)));
		cfg |= EP_STALL_CONTROL;
		iowrite32(cfg, IO_ADDRESS(EP_CONTROL + 
				(index * USBCORE_ENDPOINT_SIZE)));
	}

	return 0;
}

/**
 * clear_stall - un-stall specified endpoint
 * @ep: endpoint structure
 */
static int clear_stall(struct aspeed_ep * ep)
{
	u32 ctrl, cfg;
	int index = ep_index(ep);
	struct aspeed_request *req;

	DEBUG_ENTRY("%s, %p [ep %d]\n", __FUNCTION__, ep, index);

	if (index == 0) {
		/* EP0 */
		ctrl = ioread32(IO_ADDRESS(EP0_CONTROL));
		ctrl &= ~EP0_STALL_CONTROL;
		iowrite32(ctrl, IO_ADDRESS(EP0_CONTROL));
	} else {
		/* EP1 - EP7 */
		cfg = ioread32(IO_ADDRESS(EP_CONTROL + 
				(index * USBCORE_ENDPOINT_SIZE)));
		cfg &= ~EP_STALL_CONTROL;
		iowrite32(cfg, IO_ADDRESS(EP_CONTROL + 
				(index * USBCORE_ENDPOINT_SIZE)));
		reset_data_toggle(ep);	

		if (ep_is_in(ep)) {
			if (list_empty(&ep->queue)) 
				req = 0;
			else
				req = list_entry(ep->queue.next,
					struct aspeed_request, queue);

			if (req) {
				write_fifo(ep, req, ep->transmitBusy);
			}
		}
	}

	return 0;
}

/**
 * check_stall - check stall status for specified endpoint
 * @ep: endpoint structure
 */
static int check_stall(struct aspeed_ep * ep)
{
	u32 ctrl, cfg, stall = 0;
	int index = ep_index(ep);

	DEBUG_ENTRY("%s, %p [ep %d]\n", __FUNCTION__, ep, index);

	if (index == 0) {
		/* EP0 */
		ctrl = ioread32(IO_ADDRESS(EP0_CONTROL));
		ctrl &= EP0_STALL_CONTROL;
		if (ctrl)
			stall++;
	} else {
		/* EP1 - EP7 */
		cfg = ioread32(IO_ADDRESS(EP_CONTROL + 
				(index * USBCORE_ENDPOINT_SIZE)));
		cfg &= EP_STALL_CONTROL;
		if (cfg)
			stall++;
	}

	return stall;
}
#endif

/**
 * idle_endpoint - idle specified endpoint
 * @ep: endpoint structure
 */
static void idle_endpoint(struct aspeed_ep * ep)
{
	u32 cfg;
	int index = ep_index(ep);

	DEBUG_ENTRY("%s, %p [ep %d]\n", __FUNCTION__, ep, index);

	if (index == 0)
		return;

	iowrite32(0x4, IO_ADDRESS(EP_LIST_CONTROL +
			(index * USBCORE_ENDPOINT_SIZE)));
	iowrite32(0x2, IO_ADDRESS(EP_LIST_CONTROL +
			(index * USBCORE_ENDPOINT_SIZE)));
	udelay(1);

	iowrite32(0, IO_ADDRESS(EP_LIST_PTRS +
			(index * USBCORE_ENDPOINT_SIZE)));

	cfg = (EP_ENABLE);
	cfg |= (index << EP_NUMBER_SHFT);
	if (ep->ep_type == ep_interrupt)
		cfg |= EP_TYPE_INTERRUPT_IN;
	if (ep->ep_type == ep_bulk_in)
		cfg |= EP_TYPE_BULK_IN;
	if (ep->ep_type == ep_bulk_out)
		cfg |= EP_TYPE_BULK_OUT;
	iowrite32(cfg, IO_ADDRESS(EP_CONTROL +
		 (index * USBCORE_ENDPOINT_SIZE)));
}

/*
 * low-level aspeed 2050 usbcore transfer routines
 */

/**
 * recv_out_packet_control - read received data (OUT from host)
 * @ep: endpoint structure
 * @buf: buffer (returned data)
 * @size: size of buffer
 *
 * control endpoint specific
 * returns count of data bytes read
 */
static int recv_out_packet_control(struct aspeed_ep *ep, u8 * buf, int size)
{
	u32 val;
	int count;

	DEBUG_ENTRY("%s, %p [ep]\n", __FUNCTION__, ep);

	iowrite32(EP0_OUT_READY_TO_RECEIVE, IO_ADDRESS(EP0_CONTROL));
	val = ioread32(IO_ADDRESS(EP0_CONTROL));
	count = (val >> EP0_OUT_RECEIVED_COUNT_SHFT) & 0x7f;

	DEBUG_EP0("read packet, size %d\n", count);
	if (count) {
		if (count > size)
			count = size;
		memcpy(buf, the_controller->ep0_buffer, count);
	}

	return count;
}

/**
 * recv_out_packet_int - read received data (OUT from host)
 * @ep: endpoint structure
 * @buf: buffer (returned data)
 * @size: size of buffer
 *
 * interrupt endpoint specific
 * returns count of data bytes read
 */
static int recv_out_packet_int(struct aspeed_ep *ep, u8 * buf, int size)
{
	u32 val;
	int index = ep_index(ep);
	int count;

	DEBUG_ENTRY("%s, %p [ep %d]\n", __FUNCTION__, ep, index);

	val = ioread32(IO_ADDRESS(EP_LIST_PTRS +
		(index * USBCORE_ENDPOINT_SIZE)));
	count = (val >> 16) & 0x7ff;

	val &= ~0x1F;

	iowrite32(val,
		IO_ADDRESS(EP_LIST_PTRS +
		(index * USBCORE_ENDPOINT_SIZE)));

	if (count) {
		if (count > size)
			count = size;
		memcpy(buf, ep->buffer, count);
	}

	return count;
}

/**
 * recv_out_packet_bulk - read received data (OUT from host)
 * @ep: endpoint structure
 * @buf: buffer (returned data)
 * @size: size of buffer
 *
 * bulk endpoint specific
 * returns count of data bytes read
 */
static int recv_out_packet_bulk(struct aspeed_ep *ep, u8 * buf, int size)
{
	u32 val;
	int index = ep_index(ep);
	int count;

	DEBUG_ENTRY("%s, %p [ep %d]\n", __FUNCTION__, ep, index);

	val = ioread32(IO_ADDRESS(EP_LIST_PTRS +
		(index * USBCORE_ENDPOINT_SIZE)));
	count = (val >> 16) & 0x7ff;

	val &= ~0x1F;

	iowrite32(val,
		IO_ADDRESS(EP_LIST_PTRS +
		(index * USBCORE_ENDPOINT_SIZE)));

	if (count) {
		if (count > size)
			count = size;
		memcpy(buf, ep->buffer, count);
	}

	return count;
}

/**
 * data_in_fifo_control - return number of data bytes available
 * @ep: endpoint structure
 *
 * control endpoint specific
 * returns count of data bytes available
 */
static int data_in_fifo_control(struct aspeed_ep *ep)
{
	u32 val;
	int count;

	DEBUG_ENTRY("%s, %p [ep]\n", __FUNCTION__, ep);

	val = ioread32(IO_ADDRESS(EP0_CONTROL));
	count = (val >> EP0_OUT_RECEIVED_COUNT_SHFT) & 0x7f;

	return count;
}

/**
 * data_in_fifo_int - return number of data bytes available
 * @ep: endpoint structure
 *
 * interrupt endpoint specific
 * returns count of data bytes available
 */
static int data_in_fifo_int(struct aspeed_ep *ep)
{
	u32 val;
	int index = ep_index(ep);
	int count;

	DEBUG_ENTRY("%s, %p [ep %d]\n", __FUNCTION__, ep, index);

	val = ioread32(IO_ADDRESS(EP_LIST_PTRS +
		(index * USBCORE_ENDPOINT_SIZE)));
	count = (val >> 16) & 0x7ff;

	return count;
}

/**
 * data_in_fifo_bulk - return number of data bytes available
 * @ep: endpoint structure
 *
 * bulk endpoint specific
 * returns count of data bytes available
 */
static int data_in_fifo_bulk(struct aspeed_ep *ep)
{
	u32 val;
	int index = ep_index(ep);
	int count;

	DEBUG_ENTRY("%s, %p [ep %d]\n", __FUNCTION__, ep, index);

	val = ioread32(IO_ADDRESS(EP_LIST_PTRS +
		(index * USBCORE_ENDPOINT_SIZE)));
	count = (val >> 16) & 0x7ff;

	return count;
}

/**
 * transmit_in_packet_control - transmit data (IN to host)
 * @ep: endpoint structure
 * @buf: buffer
 * @size: amount of data
 *
 * control endpoint specific
 */
static int transmit_in_packet_control(struct aspeed_ep *ep, u8 * buf, int size)
{
	DEBUG_ENTRY("%s, %p [ep] count = %d\n", __FUNCTION__, ep, size);

	memcpy(the_controller->ep0_buffer, buf, size);
	consistent_sync(dma_to_virt(the_controller->dev, the_controller->ep0_buffer_phy), (size + LINESIZE - 1) & LINEMASK,
		DMA_TO_DEVICE);

	//dma_sync_single_for_device(the_controller->dev, the_controller->ep0_buffer_phy, (size + LINESIZE - 1) & LINEMASK,
	//	DMA_TO_DEVICE);

	iowrite32(size << EP0_IN_TRANSFER_COUNT_SHFT,
		IO_ADDRESS(EP0_CONTROL));
	iowrite32((size << EP0_IN_TRANSFER_COUNT_SHFT) |
		EP0_IN_READY_TO_TRANSFER,
		IO_ADDRESS(EP0_CONTROL));

	return size;
}

/**
 * transmit_in_packet_int - transmit data (IN to host)
 * @ep: endpoint structure
 * @buf: buffer
 * @size: amount of data
 *
 * interrupt endpoint specific
 */
static int transmit_in_packet_int(struct aspeed_ep *ep, u8 * buf, int size)
{
	int index = ep_index(ep);

	DEBUG_ENTRY("%s, %p [ep %d] count = %d\n", __FUNCTION__, ep, index, size);

	memcpy(ep->buffer, buf, size);
	consistent_sync(dma_to_virt(the_controller->dev, ep->buffer_phy), (size + LINESIZE - 1) & LINEMASK,
		DMA_TO_DEVICE);

	//dma_sync_single_for_device(the_controller->dev, ep->buffer_phy, (size + LINESIZE - 1) & LINEMASK,
	//	DMA_TO_DEVICE);

	/* endpoint send */
	iowrite32(0x4, IO_ADDRESS(EP_LIST_CONTROL +
		(index * USBCORE_ENDPOINT_SIZE)));
	iowrite32(0x2, IO_ADDRESS(EP_LIST_CONTROL +
		(index * USBCORE_ENDPOINT_SIZE)));

	iowrite32(size << 16,
		IO_ADDRESS(EP_LIST_PTRS +
		(index * USBCORE_ENDPOINT_SIZE)));
	iowrite32((size << 16) | 0x1,
		IO_ADDRESS(EP_LIST_PTRS +
		(index * USBCORE_ENDPOINT_SIZE)));

	/* set timeout for 100ms. */
	schedule_delayed_work(&ep->work, HZ/10);

	return size;
}

/**
 * transmit_in_packet_bulk - transmit data (IN to host)
 * @ep: endpoint structure
 * @buf: buffer
 * @size: amount of data
 *
 * bulk endpoint specific
 */
static int transmit_in_packet_bulk(struct aspeed_ep *ep, u8 * buf, int size)
{
	int index = ep_index(ep);

	DEBUG_ENTRY("%s, %p [ep %d] count = %d\n", __FUNCTION__, ep, index, size);

	memcpy(ep->buffer, buf, size);
	consistent_sync(dma_to_virt(the_controller->dev, ep->buffer_phy), (size + LINESIZE - 1) & LINEMASK,
		DMA_TO_DEVICE);

	//dma_sync_single_for_device(the_controller->dev, ep->buffer_phy, (size + LINESIZE - 1) & LINEMASK,
	//	DMA_TO_DEVICE);

	/* endpoint send */
	iowrite32(0x4, IO_ADDRESS(EP_LIST_CONTROL +
		(index * USBCORE_ENDPOINT_SIZE)));
	iowrite32(0x2, IO_ADDRESS(EP_LIST_CONTROL +
		(index * USBCORE_ENDPOINT_SIZE)));

	iowrite32(size << 16,
		IO_ADDRESS(EP_LIST_PTRS +
		(index * USBCORE_ENDPOINT_SIZE)));
	iowrite32((size << 16) | 0x1,
		IO_ADDRESS(EP_LIST_PTRS +
		(index * USBCORE_ENDPOINT_SIZE)));

	/* set timeout for 100ms. */
	schedule_delayed_work(&ep->work, HZ/10);

	return size;
}

/**
 * terminate_in_packet_control - terminate in transfer
 * @ep: endpoint structure
 *
 * control endpoint specific
 */
static void terminate_in_packet_control(struct aspeed_ep *ep)
{
	DEBUG_ENTRY("%s, %p [ep]\n", __FUNCTION__, ep);

	iowrite32(0, IO_ADDRESS(EP0_CONTROL));
}

/**
 * terminate_in_packet_int - terminate in transfer
 * @ep: endpoint structure
 *
 * interrupt endpoint specific
 */
static void terminate_in_packet_int(struct aspeed_ep *ep)
{
	int index = ep_index(ep);

	DEBUG_ENTRY("%s, %p [ep %d]\n", __FUNCTION__, ep, index);
	iowrite32(0x4, IO_ADDRESS(EP_LIST_CONTROL +
		(index * USBCORE_ENDPOINT_SIZE)));
	iowrite32(0x2, IO_ADDRESS(EP_LIST_CONTROL +
		(index * USBCORE_ENDPOINT_SIZE)));
	ep->transmitBusy = 0;
}

/**
 * terminate_in_packet_bulk - terminate in transfer
 * @ep: endpoint structure
 *
 * bulk endpoint specific
 */
static void terminate_in_packet_bulk(struct aspeed_ep *ep)
{
	int index = ep_index(ep);

	DEBUG_ENTRY("%s, %p [ep %d]\n", __FUNCTION__, ep, index);
	iowrite32(0x4, IO_ADDRESS(EP_LIST_CONTROL +
		(index * USBCORE_ENDPOINT_SIZE)));
	iowrite32(0x2, IO_ADDRESS(EP_LIST_CONTROL +
		(index * USBCORE_ENDPOINT_SIZE)));
	ep->transmitBusy = 0;
}

/**
 * timeout_transfer - timeout/terminate in transfer
 * @ep: endpoint structure
 */

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,23)
 static void timeout_transfer(void *_ep)
{
        struct aspeed_ep *ep = (struct aspeed_ep *) _ep;

        printk("usb_udc: timeout transfer, ep [%d]\n", ep_index(ep));

        if (ep_is_in(ep)) {
                ep->xfer_ops->terminate_in_packet(ep);
                aspeed_in_epn(ep);
        }
}
#else
static void timeout_transfer(struct work_struct *_work)
{
        struct aspeed_ep *ep;

        ep = container_of(_work, struct aspeed_ep, work.work);

        printk("usb_udc: timeout transfer, ep [%d]\n", ep_index(ep));

        if (ep_is_in(ep)) {
                ep->xfer_ops->terminate_in_packet(ep);
                aspeed_in_epn(ep);
        }
}
#endif

/**
 * ready to receive_control - set EP ready to receive
 * @ep: endpoint structure
 *
 * control endpoint specific
 */
static void ready_to_receive_control(struct aspeed_ep *ep)
{
	DEBUG_ENTRY("%s, %p [ep]\n", __FUNCTION__, ep);

	iowrite32(0, IO_ADDRESS(EP0_CONTROL));
}

/**
 * ready to receive_int - set EP ready to receive
 * @ep: endpoint structure
 *
 * interrupt endpoint specific
 */
static void ready_to_receive_int(struct aspeed_ep *ep)
{
	int index = ep_index(ep);
	u32 cfg;

	DEBUG_ENTRY("%s, %p [ep %d]\n", __FUNCTION__, ep, index);

	cfg = ioread32(IO_ADDRESS(EP_LIST_PTRS +
               	(index * USBCORE_ENDPOINT_SIZE)));
	cfg |= 0x1;
	iowrite32(cfg, IO_ADDRESS(EP_LIST_PTRS +
		(index * USBCORE_ENDPOINT_SIZE)));
}

/**
 * ready to receive_bulk - set EP ready to receive
 * @ep: endpoint structure
 *
 * bulk endpoint specific
 */
static void ready_to_receive_bulk(struct aspeed_ep *ep)
{
	int index = ep_index(ep);
	u32 cfg;

	DEBUG_ENTRY("%s, %p [ep %d]\n", __FUNCTION__, ep, index);

	cfg = ioread32(IO_ADDRESS(EP_LIST_PTRS +
               	(index * USBCORE_ENDPOINT_SIZE)));
	cfg |= 0x1;
	iowrite32(cfg, IO_ADDRESS(EP_LIST_PTRS +
		(index * USBCORE_ENDPOINT_SIZE)));
}

/*
 * EP 0 control packet routines
 */

/**
 * aspeed_handle_get_status - return device status
 * @dev:
 * @ctrl:
 */
static int aspeed_handle_get_status(struct aspeed_udc *dev,
				     struct usb_ctrlrequest *ctrl)
{
	struct aspeed_ep *ep0 = &dev->ep[0];
	struct aspeed_ep *qep;
	int reqtype = (ctrl->bRequestType & USB_RECIP_MASK);
	int cnt;
	u16 val = 0;
	u8 buf[64];

	if (reqtype == USB_RECIP_INTERFACE) {
		/* This is not supported.
		 * And according to the USB spec, this one does nothing..
		 * Just return 0
		 */
		DEBUG_EP0("GET_STATUS: USB_RECIP_INTERFACE\n");
	} else if (reqtype == USB_RECIP_DEVICE) {
		DEBUG_EP0("GET_STATUS: USB_RECIP_DEVICE\n");
		val |= (1 << 0);	/* Self powered */
	} else if (reqtype == USB_RECIP_ENDPOINT) {
		int ep_num = (ctrl->wIndex & ~USB_DIR_IN);

		DEBUG_EP0
		    ("GET_STATUS: USB_RECIP_ENDPOINT (%d), ctrl->wLength = %d\n",
		     ep_num, ctrl->wLength);

		if (ctrl->wLength > 2 || ep_num > 7)
			return -EOPNOTSUPP;

		qep = &dev->ep[ep_num];
		if (ep_is_in(qep) != ((ctrl->wIndex & USB_DIR_IN) ? 1 : 0)
		    && ep_index(qep) != 0) {
			return -EOPNOTSUPP;
		}

		/* Return status on next IN token */
#ifdef	STALL_SUPPORT
		buf[0] = check_stall(&dev->ep[ep_num]);
#else
		buf[0] = 0;
#endif
		buf[1] = 0;
		cnt = ep0->xfer_ops->recv_out_packet(ep0, buf, 2);
		memcpy(&val, buf, 2);

	} else {
		printk("usb_udc: Unknown REQ TYPE: %d\n", reqtype);
		return -EOPNOTSUPP;
	}

	/* Clear "out packet ready" */
	ep0->xfer_ops->ready_to_receive(ep0);

	/* Put status to FIFO */
	ep0->xfer_ops->transmit_in_packet(ep0, (u8 *) &val, sizeof(val));

	/* ready to receive ? */
	return 0;
}

/**
 * aspeed_ep0_setup - setup packet received (endpoint 0)
 * @dev: device structure
 */
static void aspeed_ep0_setup(struct aspeed_udc *dev)
{
	struct aspeed_ep *ep = &dev->ep[0];
	struct usb_ctrlrequest ctrl;
	int i;
	u8 buf[2];

	DEBUG_ENTRY("%s\n", __FUNCTION__);

	/* Nuke all previous transfers */
	udc_nuke(ep, -EPROTO);

	/* if the speed is not known, get it now */
	if (dev->gadget.speed == USB_SPEED_UNKNOWN) 
		dev->gadget.speed = get_speed();

	/* read control req from fifo (8 bytes) */
	read_setup_packet(dev, &ctrl);

	DEBUG_SETUP("CTRL.bRequestType = %x [%s %s] (is_in %d)\n",
		ctrl.bRequestType,
		(((ctrl.bRequestType >> 5) & 0x3) <= 0x5) ?
			typeName[(ctrl.bRequestType >> 5) & 0x3] : "unknown",
		recipientName[ctrl.bRequestType & USB_RECIP_MASK],
		ctrl.bRequestType == USB_DIR_IN);
	DEBUG_SETUP("CTRL.bRequest = %x %s\n", ctrl.bRequest,
		(((ctrl.bRequestType >> 5) & 0x3) == 0) ?
			requestName[ctrl.bRequest] : "");
	DEBUG_SETUP("CTRL.wLength = %x\n", ctrl.wLength);
	DEBUG_SETUP("CTRL.wValue = %x (%x)\n", ctrl.wValue, ctrl.wValue >> 8);
	DEBUG_SETUP("CTRL.wIndex = %x\n", ctrl.wIndex);

	/* Set direction of EP0 */
	if (likely(ctrl.bRequestType & USB_DIR_IN)) {
		ep->endpointAddress |= USB_DIR_IN;
	} else {
		ep->endpointAddress &= ~USB_DIR_IN;
	}

	/* Handle some SETUP packets ourselves */
	switch (ctrl.bRequest) {
	case USB_REQ_SET_ADDRESS:
		if (ctrl.bRequestType != (USB_TYPE_STANDARD | USB_RECIP_DEVICE))
			break;

		DEBUG_SETUP("USB_REQ_SET_ADDRESS (%d)\n", ctrl.wValue);
		device_address = ctrl.wValue;
		device_address_change = 1;
		ep->xfer_ops->transmit_in_packet(ep, buf, 0);
		return;

	case USB_REQ_GET_STATUS:
		if (aspeed_handle_get_status(dev, &ctrl) == 0)
			return;
		break;

	case USB_REQ_CLEAR_FEATURE:
	case USB_REQ_SET_FEATURE:
		if (ctrl.bRequestType == USB_RECIP_ENDPOINT) {
			struct aspeed_ep *qep;
			int ep_num = (ctrl.wIndex & 0x0f);

			/* Support only HALT feature */
			if (ctrl.wValue != 0 || ctrl.wLength != 0
			    || ep_num < 1)
				break;

			qep = &dev->ep[ep_num];

			/* Reply with a ZLP on next IN token */
			buf[0] = 0;
			buf[1] = 0;
			ep->xfer_ops->transmit_in_packet(ep, buf, 0);

			spin_unlock(&dev->lock);
			if (ctrl.bRequest == USB_REQ_SET_FEATURE) {
				DEBUG_SETUP("SET_FEATURE (%d)\n",
					    ep_num);
				aspeed_set_halt(&qep->ep, 1);
			} else {
				DEBUG_SETUP("CLR_FEATURE (%d)\n",
					    ep_num);
				aspeed_set_halt(&qep->ep, 0); 
			}
			spin_lock(&dev->lock);
			return;
		}
		break;
	
	default:
		break;
	}

    if (((ctrl.bRequestType & USB_TYPE_MASK) == USB_TYPE_STANDARD)
        && ((ctrl.bRequestType & USB_RECIP_MASK) == USB_RECIP_INTERFACE)
        && ((ctrl.bRequest == USB_REQ_SET_CONFIGURATION) 
            || (ctrl.bRequest == USB_REQ_SET_FEATURE)))
    {
        char buf[2] = {0, 0};

        ep->xfer_ops->transmit_in_packet(ep, buf, 0);
        return;
    }

	if (likely(dev->driver)) {
		spin_unlock(&dev->lock);
		i = dev->driver->setup(&dev->gadget, &ctrl);
		spin_lock(&dev->lock);

		if (i < 0) {
			/* setup processing failed, force stall */
			DEBUG_SETUP
			    ("  --> ERROR: gadget setup FAILED (stalling), setup returned %d\n",
			     i);

#ifdef	STALL_SUPPORT
			set_stall(ep);
#endif
			dev->ep0state = WAIT_FOR_SETUP;
		}
	}

	/* for HID SET_REPORT reply with ZLP on next IN token (status) */
	if (((ctrl.bRequestType & USB_TYPE_MASK) == USB_TYPE_CLASS)
	 && ((ctrl.bRequestType & USB_RECIP_MASK) == USB_RECIP_INTERFACE)
	 &&  (ctrl.wLength > 0)
	 &&  (ctrl.bRequest == HID_REQ_SET_REPORT))
		ep->xfer_ops->transmit_in_packet(ep, buf, 0);
}

/**
 * aspeed_in_epn - handle IN interrupt
 * @ep: endpoint structure
 */
static void aspeed_in_epn(struct aspeed_ep *ep)
{
	struct aspeed_request *req;
	unsigned long flags;
	struct aspeed_udc *dev = the_controller;

	DEBUG_ENTRY("%s, %p [ep %d]\n", __FUNCTION__, ep, ep_index(ep));

	if (!ep->desc) {
		DEBUG_PRINT("%s: NO EP DESC\n", __FUNCTION__);
		return;
	}

	spin_lock_irqsave(&dev->lock, flags);

	if (list_empty(&ep->queue)) 
		req = 0;
	else
		req = list_entry(ep->queue.next,
			struct aspeed_request, queue);

	if (!req) 
		ep->transmitBusy = 0;
	else	write_fifo(ep, req, ep->transmitBusy);

	spin_unlock_irqrestore(&dev->lock, flags);
}

/**
 * aspeed_out_epn - handle OUT interrupt
 * @ep: endpoint structure
 */
static void aspeed_out_epn(struct aspeed_ep *ep)
{
	struct aspeed_request *req;
	int ret, first = 1;

	DEBUG_ENTRY("%s, %p [ep %d]\n", __FUNCTION__, ep, ep_index(ep));

	if (!ep->desc) {
		DEBUG_PRINT("%s: NO EP DESC\n", __FUNCTION__);
		return;
	}

	for (;;) {
		if (list_empty(&ep->queue)) {
			req = 0;
		} else
			req = list_entry(ep->queue.next,
				struct aspeed_request, queue);

		if (!req)
			break;

		if (!first)
			ep->xfer_ops->ready_to_receive(ep);

		ret = read_fifo(ep, req);
		if (ret != 1)
			break;
		first = 0;
	}
}

/**
 * aspeed_reset_intr(dev) - handle USB reset interrupt
 * @dev: device stucture
 */
static void aspeed_reset_intr(struct aspeed_udc *dev)
{
	int i;
	struct aspeed_ep *ep;
	unsigned long flags;

	DEBUG_ENTRY("%s: %p\n", __FUNCTION__, dev);

	spin_lock_irqsave(&dev->lock, flags);

	for (i=0; i<UDC_MAX_ENDPOINTS; i++) {
		ep = &the_controller->ep[i];
		if (ep->enabled) {
			udc_nuke(ep, -ESHUTDOWN);
			idle_endpoint(ep);
		}
	}

	spin_unlock_irqrestore(&dev->lock, flags);

	if (dev->driver)
		dev->driver->disconnect(&dev->gadget);

	dev->gadget.speed = USB_SPEED_UNKNOWN;
}

/**
 * aspeed_usb_interrupt() - IRQ handler
 * @irq: IRQ value
 * @dev_id: device specific value (unused)
 * @regs: arch specific registers (unused)
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,23)
static irqreturn_t aspeed_usb_interrupt (int irq, void *dev_id, struct pt_regs *regs)
#else
static irqreturn_t aspeed_usb_interrupt (int irq, void* dev_id)
#endif
{
	u32 status;

	status = ioread32(IO_ADDRESS(INT_STATUS));

	if (status & BUS_RESET_INTR) {
		DEBUG_IRQ("%s, %s (%x)\n", 
			__FUNCTION__, "bus reset", status);
		iowrite32(BUS_RESET_INTR, IO_ADDRESS(INT_STATUS));
		aspeed_reset_intr(&udc_controller);
	}

	if (status & SUSPEND_ENTRY_INTR) {
		DEBUG_IRQ("%s, %s (%x)\n", 
			__FUNCTION__, "suspend entry", status);
		iowrite32(SUSPEND_ENTRY_INTR, IO_ADDRESS(INT_STATUS));
	}

	if (status & EP0_IN_ACK_STALL_INTR) {
		struct aspeed_ep *ep;
		struct aspeed_udc *dev;
#if 0
		u8 buf[1];
#endif

		DEBUG_IRQ("%s, %s (%x)\n", 
			__FUNCTION__, "EP0 in ack/stall", status);
		iowrite32(EP0_IN_ACK_STALL_INTR, IO_ADDRESS(INT_STATUS));

		iowrite32(0, IO_ADDRESS(EP0_CONTROL));

                ep = &udc_controller.ep[0];
	
		/* address change */
		if (device_address_change) {
			set_address(ep, device_address);
			device_address_change = 0;
		}

		dev = &udc_controller;
		if (dev->ep0state == DATA_STATE_CONTROL_READ)
			udc_ep0_in(dev, ep);
		if (dev->ep0state == DATA_STATE_CONTROL_READ_ZLP) {
#if 0
			ep->xfer_ops->transmit_in_packet(ep, buf, 0);
#endif
			dev->ep0state = WAIT_FOR_SETUP;
		}
	}

	if (status & EP0_OUT_ACK_STALL_INTR) {
		DEBUG_IRQ("%s, %s (%x)\n", 
			__FUNCTION__, "EP0 out ack/stall", status);
		iowrite32(EP0_OUT_ACK_STALL_INTR, IO_ADDRESS(INT_STATUS));
	}

	if (status & EP0_OUT_NAK_INTR) {
		struct aspeed_ep *ep;
		DEBUG_IRQ("%s, %s (%x)\n", 
			__FUNCTION__, "EP0 out nak", status);
		iowrite32(EP0_OUT_NAK_INTR, IO_ADDRESS(INT_STATUS));
                ep = &udc_controller.ep[0];
		udc_ep0_out(&udc_controller, ep);
	}

	if (status & SUSPEND_RESUME_INTR) {
		DEBUG_IRQ("%s, %s (%x)\n", 
			__FUNCTION__, "suspend resume", status);
		iowrite32(SUSPEND_RESUME_INTR, IO_ADDRESS(INT_STATUS));
	}

	if (status & BUS_DEADLOCKED) { 
		DEBUG_IRQ("%s, %s (%x)\n", 
			__FUNCTION__, "bus deadlocked", status);
		iowrite32(BUS_DEADLOCKED, IO_ADDRESS(INT_STATUS));
	}

	if (status & ENDPOINT_POOL_ACK_STALL_INTR) {
		u32 ack;
		int index;
		struct aspeed_ep *ep;

		DEBUG_IRQ("%s, %s (%x)\n", 
			__FUNCTION__, "endpool ack/stall", status);
		iowrite32(ENDPOINT_POOL_ACK_STALL_INTR, IO_ADDRESS(INT_STATUS));
		ack = ioread32(IO_ADDRESS(EP_ACK_INT_STATUS));
		iowrite32(ack, IO_ADDRESS(EP_ACK_INT_STATUS));

		index = 0;
		ack &= 0xff;
		while (ack) {
			if (ack & 0x1) {
				ep = &udc_controller.ep[index];
				cancel_delayed_work(&ep->work);
				if (ep_is_in(ep))	
					aspeed_in_epn(ep);
				else	aspeed_out_epn(ep);

			}
			index++;
			ack = ack >> 1;
		}
	}

	if (status & EP0_SETUP_ACK_INTR) {
		DEBUG_IRQ("%s, %s (%x)\n", 
			__FUNCTION__, "EP0 setup ack", status);
		iowrite32(EP0_SETUP_ACK_INTR, IO_ADDRESS(INT_STATUS));
		aspeed_ep0_setup(&udc_controller);
	}

	return IRQ_HANDLED;
}

/**
 * usbcore_reset() - reset usb core
 */
static void usbcore_reset(void)
{
	u32 reset;

	reset = ioread32(IO_ADDRESS(SCU_RESET_CONTROL));
	reset |= RESET_USB20;
        iowrite32(reset, IO_ADDRESS(SCU_RESET_CONTROL));
	reset &= ~RESET_USB20;
        iowrite32(reset, IO_ADDRESS(SCU_RESET_CONTROL));

	iowrite32((PHYCLOCK | ENABLE_PHY_RESET),
		IO_ADDRESS(ROOT_CONTROL));
	udelay(1);

	iowrite32(0, IO_ADDRESS(SW_RESET));
	iowrite32(0, IO_ADDRESS(INT_CONTROL));
	iowrite32(0, IO_ADDRESS(INT_STATUS));
}

/**
 * usbcore_start() - reset and start usb core
 */
static void usbcore_start(void)
{
	usbcore_reset();

	iowrite32(0, IO_ADDRESS(SW_RESET));
	iowrite32(INT_CONTROL_INIT, IO_ADDRESS(INT_CONTROL));
	iowrite32(INT_STATUS_INIT, IO_ADDRESS(INT_STATUS));

	iowrite32(0xff, IO_ADDRESS(EP_ACK_INT_CONTROL));
	mdelay(1);
	iowrite32(0x0, IO_ADDRESS(EP_NAK_INT_CONTROL));
	mdelay(1);
	iowrite32(0xffffff, IO_ADDRESS(EP_ACK_INT_STATUS));
	mdelay(1);
	iowrite32(0xffffff, IO_ADDRESS(EP_NAK_INT_STATUS));
	mdelay(1);

	iowrite32(0x0, IO_ADDRESS(EP0_CONTROL));
	iowrite32(0, IO_ADDRESS(EP1_CONTROL));
	udelay(1);

	iowrite32((PHYCLOCK | ENABLE_PHY_RESET |
		ENABLE_UPSTREAM_PORT_CONNECTION),
		IO_ADDRESS(ROOT_CONTROL));
	udelay(1);

        iowrite32(udc_controller.ep0_buffer_phy, IO_ADDRESS(EP0_BASE));

	udc_controller.core_initialized++;
}

/**
 * usbcore_init() - attach IRQ
 *
 * Fails if cannot attach to IRQ.
 */
static int __devinit usbcore_init(void)
{
	int retval;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,23)
        u32                            irqFlags = SA_INTERRUPT;
#else
        u32                            irqFlags = IRQF_DISABLED | IRQF_SAMPLE_RANDOM;
#endif

	usbcore_clock_start();
	usbcore_reset();

        IRQ_SET_HIGH_LEVEL (USBCORE_IRQ);
        IRQ_SET_LEVEL_TRIGGER (USBCORE_IRQ);
        IRQ_EDGE_CLEAR(USBCORE_IRQ);
        iowrite32(0x1ff, IO_ADDRESS(INT_STATUS));

	retval = request_irq (USBCORE_IRQ, &aspeed_usb_interrupt,
		irqFlags, "aspeed_usb20", NULL);
	if (retval != 0) {
		printk(KERN_ERR "%s: can't get irq %i, err %d\n", driver_name,
		      USBCORE_IRQ, retval);
		return -EBUSY;
	}

	return 0;
}

/**
 * usbdata_init() - allocate buffers
 */
static int __devinit usbdata_init(void)
{
	void * buf;
	dma_addr_t dma;
	int i;
	struct aspeed_ep *ep;

	spin_lock_init(&udc_controller.lock);

	udc_controller.setup_low = (u32 *) IO_ADDRESS(ROOT_SETUP_LOW);
	udc_controller.setup_high = (u32 *) IO_ADDRESS(ROOT_SETUP_HIGH);

	/* endpoint 0 buffer */
        buf = dma_alloc_coherent(udc_controller.gadget.dev.parent,
		64, &dma, GFP_DMA | GFP_KERNEL);
	if (buf == (void *) NULL)
		return -ENOMEM;
	if ((u32)buf & 0x7)
		return -ENOMEM;

	udc_controller.ep0_buffer = buf;
	udc_controller.ep0_buffer_phy = dma;
	udc_controller.control_buf_count = 0;

	/* buffer descriptors */
	for (i=0; i<UDC_MAX_ENDPOINTS; i++) {
		ep = &udc_controller.ep[i];
        	buf = dma_alloc_coherent(udc_controller.gadget.dev.parent,
			1024, &dma, GFP_DMA | GFP_KERNEL);
		if (buf == (void *) NULL) 
			return -ENOMEM;
		if ((u32)buf & 0x7) 
			return -ENOMEM;

		ep->buffer = buf;	
		ep->buffer_phy = dma;
	}

	return 0;
}

/**
 * udc_init() - module load
 */
static int __devinit udc_init(void)
{
	int retval, i;
	struct aspeed_ep *ep;

	DEBUG_ENTRY("%s: %s version %s\n", __FUNCTION__, driver_name, DRIVER_VERSION);

	retval = usbdata_init();
	if (retval != 0)
		return retval;
	retval = usbcore_init();
	if (retval != 0)
		return retval;

	for (i=0; i<UDC_MAX_ENDPOINTS; i++) {
		ep = &udc_controller.ep[i];
		ep->enabled = 0;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,23)
                INIT_WORK(&ep->work, timeout_transfer, ep);
#else
                INIT_DELAYED_WORK(&(ep->work), timeout_transfer);
#endif
	}

	platform_device_register(&aspeed_usb_gadget_device);
	return driver_register(&udc_driver);
}

/**
 * usbcore_stop() - disable USB core
 */
static void usbcore_stop(void)
{
	iowrite32(0xff, IO_ADDRESS(EP_ACK_INT_CONTROL));
	mdelay(1);

	udc_controller.core_initialized = 0;
}

/**
 * usbcore_clock_start() - start USB clock
 */
static void usbcore_clock_start(void)
{
	u32 clock;

	iowrite32(PROTECTION_KEY_UNLOCK, IO_ADDRESS(SCU_PROTECTION_KEY));

	clock = ioread32(IO_ADDRESS(SCU_CLOCK_CONTROL));
	clock |= ENABLE_USB20_CLOCK;
        iowrite32(clock, IO_ADDRESS(SCU_CLOCK_CONTROL));
	mdelay(10);
}

/**
 * usbcore_clock_stop() - stop USB clock
 */
static void usbcore_clock_stop(void)
{
	u32 clock;

	clock = ioread32(IO_ADDRESS(SCU_CLOCK_CONTROL));
	clock &= ~ENABLE_USB20_CLOCK;
        iowrite32(clock, IO_ADDRESS(SCU_CLOCK_CONTROL));
}

/**
 * usbcode_exit() - disable USB core and release IRQ
 */
static int __devexit usbcore_exit(void)
{
	usbcore_stop();
	usbcore_clock_stop();
	free_irq(USBCORE_IRQ, NULL);
}

/**
 * usbdata_free() - release allocated buffers
 */
static void __devinit usbdata_free(void)
{
	int	i;

	dma_free_coherent(the_controller->gadget.dev.parent,
		64, udc_controller.ep0_buffer,
		udc_controller.ep0_buffer_phy);

	for (i=0; i<UDC_MAX_ENDPOINTS; i++) {
		dma_free_coherent(udc_controller.gadget.dev.parent,
			1024, udc_controller.ep[i].buffer,
			udc_controller.ep[i].buffer_phy);
	}
}

/**
 * udc_exit() - module unload
 */
static void __devexit udc_exit(void)
{
	DEBUG_ENTRY("%s\n", __FUNCTION__);

	usbcore_exit();
	usbdata_free();

	platform_device_unregister(&aspeed_usb_gadget_device);
	driver_unregister(&udc_driver);
}

module_init(udc_init);
module_exit(udc_exit);

MODULE_DESCRIPTION(DRIVER_DESC);
MODULE_LICENSE("GPL");
