/*
 * g_kbdmouse.c -- Gadget HIS keyboard and mouse driver
 *
 * Copyright (C) 2006-2007 Renesas Solutions Corp.
 *
 * This code is based in part on the Gadget Zero driver, which
 * is Copyright (C) 2003-2004 by David Brownell, all rights reserved.
 * 
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 */
#ifndef LINUX_VERSION_CODE
#  include <linux/version.h>
#endif

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/delay.h>
#include <linux/ioport.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/smp_lock.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/timer.h>
#include <linux/list.h>
#include <linux/interrupt.h>
#include <linux/uts.h>
#include <linux/version.h>
#include <linux/device.h>
#include <linux/wait.h>
#include <linux/dma-mapping.h>

#include <asm/byteorder.h>
//#include <asm/io.h>
#include <asm/irq.h>
#include <asm/system.h>
#include <asm/unaligned.h>

#include <asm/uaccess.h>

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,21)
   #include <linux/config.h>
   #include <linux/usb_ch9.h>
#else
   #include <linux/usb/ch9.h>
#endif

#include <linux/usb_gadget.h>
#include <linux/fs.h>
#include <linux/cdev.h>

#include "gadget_chips.h"

#include <linux/input.h>
#include "composite.h"
#include "ch11.h"

/*-------------------------------------------------------------------------*/
#define DRIVER_DESC		"HUB Function"
#define DRIVER_NAME		"g_hub"
#define DRIVER_VERSION		"20090818"
#define CONTROLLER_NAME         "PCD"

static const char shortname [] = DRIVER_NAME;
static const char longname [] = DRIVER_DESC;

/*-------------------------------------------------------------------------*/

static struct {
	char            *controller;
} mod_data = {
	.controller             = CONTROLLER_NAME,
};

static char controller[16];

module_param_named(controller, mod_data.controller, charp, S_IRUGO);
MODULE_PARM_DESC(controller, "Name of peripheral controller driver");


/*-------------------------------------------------------------------------*/
/*
 * driver assumes self-powered hardware, and
 * has no way for users to trigger remote wakeup.
 *
 * this version autoconfigures as much as possible,
 * which is reasonable for most "bulk-only" drivers.
 */


/*-------------------------------------------------------------------------*/

/* big enough to hold our biggest descriptor */
#define USB_BUFSIZ	256

struct hub_dev {
	spinlock_t		lock;
	struct usb_gadget	*gadget;
	struct usb_request	*req;		/* for control responses */
	/* autoresume timer */
	struct timer_list       resume;
	u8			config;
};

#ifdef CONFIG_USB_COMPOSITE
int function_index = 0;
#endif

/*-------------------------------------------------------------------------*/
static struct hub_dev *g_hub;

/*
 * if it's nonzero, autoresume says how many seconds to wait
 * before trying to wake up the host after suspend.
 */
static unsigned autoresume = 0;

/*-------------------------------------------------------------------------*/

/*
 * DESCRIPTORS ... most are static, but strings and (full)
 * configuration descriptors are built on demand.
 */
#define STRING_STATUS_INTERFACE	1

#ifndef CONFIG_USB_COMPOSITE
  #define DRIVER_VENDOR_ID        0x0624	// Avocent
  #define DRIVER_PRODUCT_ID       0x0248
  
  #define STRING_MANUFACTURER			4
  #define STRING_PRODUCT			5
  #define STRING_SERIAL				6

static struct usb_device_descriptor device_desc = {
	.bLength                = sizeof device_desc,
	.bDescriptorType        = USB_DT_DEVICE,

	.bcdUSB                 = __constant_cpu_to_le16 (0x0200),
	.bDeviceClass           = USB_CLASS_HUB,
	.bDeviceSubClass        = 0,
	.bDeviceProtocol        = 1, // Single-TT

	.idVendor               = __constant_cpu_to_le16 (DRIVER_VENDOR_ID),
	.idProduct              = __constant_cpu_to_le16 (DRIVER_PRODUCT_ID),
	.iManufacturer          = STRING_MANUFACTURER,
	.iProduct               = STRING_PRODUCT,
	.iSerialNumber          = STRING_SERIAL,
	.bNumConfigurations     = 1,
};

static struct usb_config_descriptor hid_config = {
	.bLength                = sizeof hid_config,
	.bDescriptorType        = USB_DT_CONFIG,

	/* wTotalLength computed by usb_gadget_config_buf() */
	.bNumInterfaces         = 1,
	.bConfigurationValue    = 1,
	.iConfiguration         = 0,
	.bmAttributes           = USB_CONFIG_ATT_ONE | USB_CONFIG_ATT_SELFPOWER | USB_CONFIG_ATT_WAKEUP,
	.bMaxPower              = 50,	/* self-powered */
};

static struct usb_qualifier_descriptor dev_qualifier = {
	.bLength                = sizeof dev_qualifier,
	.bDescriptorType        = USB_DT_DEVICE_QUALIFIER,
	
	.bcdUSB                 = __constant_cpu_to_le16 (0x0200),	
	.bDeviceClass           = USB_CLASS_HUB,
	.bDeviceProtocol        = 0,
	.bMaxPacketSize0        = 0,
	.bNumConfigurations     = 1,
};

#endif

static struct usb_otg_descriptor otg_desc =
{
	.bLength =		sizeof(otg_desc),
	.bDescriptorType =	USB_DT_OTG,

	.bmAttributes =		USB_OTG_SRP,
};


static struct usb_hub_descriptor hub_desc =
{
	.bDescLength = 		sizeof (struct usb_hub_descriptor),
	.bDescriptorType =	USB_DT_HUB,
	.bNbrPorts =		ASPEED_NUM_PORTS,
	.wHubCharacteristics =	__constant_cpu_to_le16 (0x00E0),
	.bPwrOn2PwrGood =	0x32,
	.bHubContrCurrent =	0x64,
	.DeviceRemovable =	{0x00},
	.PortPwrCtrlMask =	{0xFF},
};

/* one interface in each configuration */
static struct usb_interface_descriptor intf_desc =
{
	.bLength =		sizeof (struct usb_interface_descriptor),
	.bDescriptorType =	USB_DT_INTERFACE,

	.bInterfaceNumber =	0,
	.bNumEndpoints =	1,
	.bInterfaceClass =	USB_CLASS_HUB,
	.bInterfaceSubClass =	0,
	.bInterfaceProtocol =	0,
	.iInterface =		STRING_STATUS_INTERFACE,
};

#ifdef CONFIG_USB_COMPOSITE
static struct usb_interface_info intf_info =
{
	.enabled		= 1,
	.intf_desc		= &intf_desc,
};

static struct usb_interface_info *intf_info_table[] = {
	&intf_info,
	NULL,
};
#endif

static struct usb_endpoint_descriptor fs_desc =
{
	.bLength =		USB_DT_ENDPOINT_SIZE,
	.bDescriptorType =	USB_DT_ENDPOINT,

	.bEndpointAddress =	0x01 | USB_DIR_IN,
	.bmAttributes =		USB_ENDPOINT_XFER_INT,
	.wMaxPacketSize =	__constant_cpu_to_le16 (1),
	.bInterval =		0x0C,
};

static const struct usb_descriptor_header *fs_hub_function [] = {
	(struct usb_descriptor_header *) &otg_desc,
	(struct usb_descriptor_header *) &intf_desc,
	(struct usb_descriptor_header *) &fs_desc,
	NULL,
};

#ifdef	CONFIG_USB_GADGET_DUALSPEED

/*
 * usb 2.0 devices need to expose both high speed and full speed
 * descriptors, unless they only run at full speed.
 *
 * that means alternate endpoint descriptors (bigger packets)
 * and a "device qualifier" ... plus more construction options
 * for the config descriptor.
 */

static struct usb_endpoint_descriptor hs_desc =
{
	.bLength =		USB_DT_ENDPOINT_SIZE,
	.bDescriptorType =	USB_DT_ENDPOINT,

	.bEndpointAddress =	0x01 | USB_DIR_IN,
	.bmAttributes =		USB_ENDPOINT_XFER_INT,
	.wMaxPacketSize =	__constant_cpu_to_le16 (1),
	.bInterval =		0x0C,
};

static const struct usb_descriptor_header *hs_hub_function [] = {
	(struct usb_descriptor_header *) &otg_desc,
	(struct usb_descriptor_header *) &intf_desc,
	(struct usb_descriptor_header *) &hs_desc,
	NULL,
};

/* maxpacket and other transfer characteristics vary by speed. */
#define ep_desc(g,hs,fs) (((g)->speed==USB_SPEED_HIGH)?(hs):(fs))

#else

/* if there's no high speed support, maxpacket doesn't change. */
#define ep_desc(g,hs,fs) fs

#endif	/* !CONFIG_USB_GADGET_DUALSPEED */

/* static strings, in iso 8859/1 */
static struct usb_string		strings [] = {
	{STRING_STATUS_INTERFACE,	"USB HUB Status"},
#ifndef CONFIG_USB_COMPOSITE
	{ STRING_MANUFACTURER,		"Avocent",},
	{ STRING_PRODUCT,		longname,},
	{ STRING_SERIAL,		DRIVER_VERSION,},
#endif
	{  }			/* end of list */
};

static struct usb_gadget_strings	stringtab = {
	.language	= 0x0409,	/* en-us */
	.strings	= strings,
};

#ifdef CONFIG_USB_COMPOSITE
static const struct usb_gadget_strings *stringTable[] = {
	(struct usb_gadget_strings *) &stringtab,
	NULL,
};
#endif

#ifndef CONFIG_USB_COMPOSITE
/*
 * config descriptors are also handcrafted.  these must agree with code
 * that sets configurations, and with code managing interfaces and their
 * altsettings.  other complexity may come from:
 *
 *  - high speed support, including "other speed config" rules
 *  - multiple configurations
 *  - interfaces with alternate settings
 *  - embedded class or vendor-specific descriptors
 *
 * this handles high speed, and has a second config that could as easily
 * have been an alternate interface setting (on most hardware).
 *
 * NOTE:  to demonstrate (and test) more USB capabilities, this driver
 * should include an altsetting to test interrupt transfers, including
 * high bandwidth modes at high speed.  (Maybe work like Intel's test
 * device?)
 */
static int config_buf (struct usb_gadget *gadget,
		u8 *buf, u8 type, unsigned index)
{
	int				   len;
	const struct usb_descriptor_header **function;
#ifdef CONFIG_USB_GADGET_DUALSPEED
	int				hs = (gadget->speed == USB_SPEED_HIGH);
#endif

	/* two configurations will always be index 0 and index 1 */
	if (index > 1)
		return -EINVAL;

#ifdef CONFIG_USB_GADGET_DUALSPEED
	if (type == USB_DT_OTHER_SPEED_CONFIG)
		hs = !hs;
	if (hs)
		function = hs_hub_function;
	else
#endif
		function = fs_hub_function;
		
	if (!gadget->is_otg)
		function++;

	len = usb_gadget_config_buf (&hid_config, buf, USB_BUFSIZ, function);
	if (len < 0)
		return len;
	((struct usb_config_descriptor *) buf)->bDescriptorType = type;
	return len;
}
#endif

/*-------------------------------------------------------------------------*/

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,23)
void * _usb_ep_alloc_buffer(struct usb_gadget *gadget,
				   struct usb_ep *_ep,
                                   unsigned bytes,
                                   dma_addr_t *dma,
                                   gfp_t gfp_flags)
{
	void *buf;
	
	*dma = DMA_ADDR_INVALID;	
	buf = dma_alloc_coherent(gadget->dev.parent,
				bytes, dma, gfp_flags);

	return buf;	
}

void _usb_ep_free_buffer(struct usb_gadget *gadget,
				struct usb_ep *_ep,
                                void *buf,
                                dma_addr_t dma,
                                unsigned bytes)
{	
	if (dma != DMA_ADDR_INVALID) {
		dma_free_coherent(gadget->dev.parent,
					bytes, buf, dma);
	}	
}
#endif

/*-------------------------------------------------------------------------*/

static void hub_reset_config (struct hub_dev *dev)
{
	if (dev->config == 0)
		return;

	//printk(KERN_DEBUG "%s: reset config\n", __FUNCTION__);

	dev->config = 0;
	del_timer (&dev->resume);
}

/* change our operational config.  this code must agree with the code
 * that returns config descriptors, and altsetting code.
 *
 * it's also responsible for power management interactions. some
 * configurations might not work with our current power sources.
 *
 * note that some device controller hardware will constrain what this
 * code can do, perhaps by disallowing more than one configuration or
 * by limiting configuration choices (like the pxa2xx).
 */
static int hub_set_config (struct hub_dev *dev, unsigned number, int gfp_flags)
{
	int			result = 0;
	struct usb_gadget	*gadget = dev->gadget;
	char                    *speed;

	if (number == dev->config)
		return 0;

	hub_reset_config (dev);

	// Don't care what configuration is
	if (number == 0)
		return result;

	switch (gadget->speed) {
	case USB_SPEED_LOW:	speed = "low"; break;
	case USB_SPEED_FULL:	speed = "full"; break;
	case USB_SPEED_HIGH:	speed = "high"; break;
	default: 		speed = "?"; break;
	}

	dev->config = number;
		
	printk(KERN_ALERT DRIVER_NAME ": %s speed config #%d"
				" for %s\n", speed, number,
				"USB Hub");		
	return result;
}

static bool isValidHubFeature(u16 value)
{
        switch (value) {
	case C_HUB_LOCAL_POWER:
	case C_HUB_OVER_CURRENT:
                return true;
        default:
                printk(KERN_ERR "%s: Unexpected HUB Feature Selector %d\n",
                                        __FUNCTION__, value);
                return false;
        }
}

static bool isValidPortFeature(u16 value)
{
	switch (value) {
	case USB_PORT_FEAT_CONNECTION:
	case USB_PORT_FEAT_ENABLE:
	case USB_PORT_FEAT_SUSPEND:
	case USB_PORT_FEAT_OVER_CURRENT:
	case USB_PORT_FEAT_RESET:
	case USB_PORT_FEAT_POWER:
	case USB_PORT_FEAT_LOWSPEED:
	case USB_PORT_FEAT_HIGHSPEED:
	case USB_PORT_FEAT_C_CONNECTION:
	case USB_PORT_FEAT_C_ENABLE:
	case USB_PORT_FEAT_C_SUSPEND:
	case USB_PORT_FEAT_C_OVER_CURRENT:
	case USB_PORT_FEAT_C_RESET:
	case USB_PORT_FEAT_TEST:
	case USB_PORT_FEAT_INDICATOR:
		return true;
	default:
		printk(KERN_ERR "%s: Unexpected Port Feature Selector %d\n",
					__FUNCTION__, value);
		return false;
	}
}

#ifdef CONFIG_USB_COMPOSITE
static int hub_class_request (struct usb_composite_dev *cdev, const struct usb_ctrlrequest *ctrl)
#else
static int hub_class_request (struct usb_gadget *gadget, const struct usb_ctrlrequest *ctrl)
#endif
{
#ifdef CONFIG_USB_COMPOSITE
	struct hub_dev		*dev = get_composite_data(cdev, function_index);
#else
	struct hub_dev		*dev = get_gadget_data(gadget);
#endif
	struct usb_request	*req = dev->req;
	u16			w_index = le16_to_cpu(ctrl->wIndex);
	u16			w_value = le16_to_cpu(ctrl->wValue);
	u16			w_length = le16_to_cpu(ctrl->wLength); 
	int                     value = -EOPNOTSUPP;
	struct usb_port_status  portStatus;

	switch (ctrl->bRequest) {
	case USB_REQ_GET_STATUS:
		if ((ctrl->bRequestType & USB_RECIP_MASK) == USB_RECIP_DEVICE) {
			//printk(KERN_DEBUG "%s(): Received GET_HUB_STATUS\n", __FUNCTION__);
			if (w_value != 0 || w_index != 0 || w_length != 4)
				break;
		} else if ((ctrl->bRequestType & USB_RECIP_MASK) == USB_RECIP_OTHER) {
			//printk(KERN_DEBUG "%s(): Received GET_PORT_STATUS\n", __FUNCTION__);
			if (w_value != 0 || w_index < 1 || w_index > ASPEED_NUM_PORTS || w_length != 4)
				break;
			{
			hub_get_port_status(w_index, &portStatus);
			value = min (ctrl->wLength, (u16) sizeof(portStatus));
			memcpy (req->buf, &portStatus, value);
			}
		}
		break;
	case USB_REQ_CLEAR_FEATURE:
		if ((ctrl->bRequestType & USB_RECIP_MASK) == USB_RECIP_DEVICE) {
			//printk(KERN_DEBUG "%s(): Received CLEAR_HUB_FEATURE\n", __FUNCTION__);
			if (!isValidHubFeature(w_value) || w_index != 0 || w_length != 0)
				break;
		} else if ((ctrl->bRequestType & USB_RECIP_MASK) == USB_RECIP_OTHER) {
			//printk(KERN_DEBUG "%s(): Received CLEAR_PORT_FEATURE\n", __FUNCTION__);
			if (!isValidPortFeature(w_value) || ((w_index & 0xFF) < 1) ||
					((w_index & 0xFF) > ASPEED_NUM_PORTS) || w_length != 0)
				break;
			if (w_value == 0x01)
				value = hub_set_clear_port_feature((w_index & 0xFF), USB_PORT_STAT_ENABLE, 0, 1);
			else if (w_value == 0x14)
				value = hub_set_clear_port_feature((w_index & 0xFF), USB_PORT_STAT_C_RESET, 0, 0);
			else if (w_value == 0x10)
				value = hub_set_clear_port_feature((w_index & 0xFF), USB_PORT_STAT_C_CONNECTION, 0, 0);
		}
		value = 0;
		break;
	case USB_REQ_SET_FEATURE:
		if ((ctrl->bRequestType & USB_RECIP_MASK) == USB_RECIP_DEVICE) {
			//printk(KERN_DEBUG "%s(): Received SET_HUB_FEATURE\n", __FUNCTION__);
			if (!isValidHubFeature(w_value) || w_index != 0 || w_length != 0)
				break;
		} else if ((ctrl->bRequestType & USB_RECIP_MASK) == USB_RECIP_OTHER) {
			//printk(KERN_DEBUG "%s(): Received SET_PORT_FEATURE\n", __FUNCTION__);
			if (!isValidPortFeature(w_value) || ((w_index & 0xFF) < 1) ||
					((w_index & 0xFF) > ASPEED_NUM_PORTS) || w_length != 0)
				break;
			if (w_value == 0x04)
				value = hub_set_clear_port_feature((w_index & 0xFF), USB_PORT_STAT_RESET, 1, 1);
			else if (w_value == 0x08)
				value = hub_set_clear_port_feature((w_index & 0xFF), USB_PORT_STAT_POWER, 1, 1);
		}
		value = 0;
		break;
	case USB_REQ_GET_DESCRIPTOR:
		//printk(KERN_DEBUG "%s: Received HUB_REQ_GET_DESCRIPTOR\n", __FUNCTION__);
		if (w_value != 0 || w_index != 0)
			break;
		value = min(ctrl->wLength, (u16) sizeof hub_desc);
		memcpy (req->buf, &hub_desc, value);
		break;
	case USB_REQ_SET_DESCRIPTOR:
		//printk(KERN_DEBUG "%s: Received HUB_REQ_SET_DESCRIPTOR\n", __FUNCTION__);
		if (w_value != 0 || w_index != 0)
			break;
		break;
	case HUB_CLEAR_TT_BUFFER:
		//printk(KERN_DEBUG "%s: Received HUB_CLEAR_TT_BUFFER\n", __FUNCTION__);
		if (w_index != 1 || w_length != 0)
			break;
		break;
	case HUB_RESET_TT:
		//printk(KERN_DEBUG "%s: Received HUB_RESET_TT\n", __FUNCTION__);
		if (w_index != 1 || w_length != 0)
			break;
		break;
	case HUB_GET_TT_STATE:
		//printk(KERN_DEBUG "%s: Received HUB_GET_TT_STATE\n", __FUNCTION__);
		if (w_index != 1)
			break;
		break;
	case HUB_STOP_TT:
		//printk(KERN_DEBUG "%s: Received HUB_STOP_TT\n", __FUNCTION__);
		if (w_index != 1 || w_length != 0)
			break;
		break;
	default:
		printk(KERN_ERR "%s: unexpected class request %x\n",
					__FUNCTION__, ctrl->bRequest);
		break;
	}

	return value;
}

static void hub_setup_complete (struct usb_ep *ep, struct usb_request *req)
{
	if (req->status || req->actual != req->length) {
		printk(KERN_ALERT "%s: %d, %d/%d\n", __FUNCTION__,
				req->status, req->actual, req->length);
	}
}

/*
 * The setup() callback implements all the ep0 functionality that's
 * not handled lower down, in hardware or the hardware driver (like
 * device and endpoint feature flags, and their status).  It's all
 * housekeeping for the gadget function we're implementing.  Most of
 * the work is in config-specific setup.
 */
#ifdef CONFIG_USB_COMPOSITE
static int hub_setup (struct usb_composite_dev *cdev, const struct usb_ctrlrequest *ctrl)
#else
static int hub_setup (struct usb_gadget *gadget, const struct usb_ctrlrequest *ctrl)
#endif
{
#ifdef CONFIG_USB_COMPOSITE
	struct hub_dev		*dev = get_composite_data(cdev, function_index);
	struct usb_gadget	*gadget = dev->gadget;
#else
	struct hub_dev		*dev = get_gadget_data(gadget);
#endif
	struct usb_request	*req = dev->req;
	int			value = -EOPNOTSUPP;
	u8			respond = 1;
	u16			w_index = le16_to_cpu(ctrl->wIndex);
	u16			w_value = le16_to_cpu(ctrl->wValue);
	u16			w_length = le16_to_cpu(ctrl->wLength);

	/* usually this stores reply data in the pre-allocated ep0 buffer,
	 * but config change events will reconfigure hardware.
	 */
	req->zero = 0;

	if ((ctrl->bRequestType & USB_TYPE_MASK) == USB_TYPE_STANDARD) {
		switch (ctrl->bRequest) {
		case USB_REQ_GET_DESCRIPTOR:
			if (!(ctrl->bRequestType & USB_DIR_IN))
				goto unknown;
			switch (w_value >> 8) {
#ifndef CONFIG_USB_COMPOSITE
			case USB_DT_DEVICE:
				value = min (ctrl->wLength, (u16) sizeof device_desc);
				memcpy (req->buf, &device_desc, value);
				break;
#ifdef CONFIG_USB_GADGET_DUALSPEED
			case USB_DT_DEVICE_QUALIFIER:
				if (!gadget->is_dualspeed)
					break;
				value = min (ctrl->wLength, (u16) sizeof dev_qualifier);
				memcpy (req->buf, &dev_qualifier, value);
				break;
			case USB_DT_OTHER_SPEED_CONFIG:
				if (!gadget->is_dualspeed)
					break;
				// FALLTHROUGH
#endif /* CONFIG_USB_GADGET_DUALSPEED */
			case USB_DT_CONFIG:
				value = config_buf (gadget, req->buf,
						ctrl->wValue >> 8,
						ctrl->wValue & 0xff);
				if (value >= 0)
					value = min (ctrl->wLength, (u16) value);
				break;
#endif
#ifndef CONFIG_USB_COMPOSITE
			case USB_DT_STRING:
				/* wIndex == language code.
	 			 * this driver only handles one language, you can
				 * add others even if they don't use iso8859/1
				 */
				value = usb_gadget_get_string (&stringtab,
						ctrl->wValue & 0xff, req->buf);
				if (value >= 0)
					value = min (ctrl->wLength, (u16) value);
				break;
#endif
			default:
				break;
			}
				break;
		/* currently two configs, two speeds */
		case USB_REQ_SET_CONFIGURATION:
			if (ctrl->bRequestType != 0)
				goto unknown;
			spin_lock (&dev->lock);
			value = hub_set_config (dev, w_value, GFP_ATOMIC);
			spin_unlock (&dev->lock);
#ifdef CONFIG_USB_COMPOSITE
			respond = 0;  // composite layer will respond to this
#endif
			break;
		case USB_REQ_GET_CONFIGURATION:
			if (ctrl->bRequestType != USB_DIR_IN)
				goto unknown;
			*(u8 *)req->buf = dev->config;
			value = min (w_length, (u16) 1);
			break;
		case USB_REQ_SET_INTERFACE:
			if (ctrl->bRequestType != USB_RECIP_INTERFACE)
				goto unknown;
			spin_lock (&dev->lock);
			if (dev->config && w_index == 0 && w_value == 0) {
				u8 config = dev->config;

				/* resets interface configuration, forgets about
				 * previous transaction state (queued bufs, etc)
				 * and re-inits endpoint state (toggle etc)
				 * no response queued, just zero status == success.
				 * if we had more than one interface we couldn't
				 * use this "reset the config" shortcut.
				 */
				hub_reset_config (dev);
				hub_set_config (dev, config, GFP_ATOMIC);
				value = 0;
			}
			spin_unlock (&dev->lock);
			break;
		case USB_REQ_GET_INTERFACE:
			if (ctrl->bRequestType != (USB_DIR_IN|USB_RECIP_INTERFACE))
				goto unknown;
			if (!dev->config)
				break;
			if (w_index != 0) {
				value = -EDOM;
				break;
			}
			*(u8 *)req->buf = 0;
			value = min (w_length, (u16) 1);
			break;

		default:
			goto unknown;
		}
	} else if ((ctrl->bRequestType & USB_TYPE_MASK) == USB_TYPE_CLASS) {
#ifdef CONFIG_USB_COMPOSITE		
		value = hub_class_request(cdev, ctrl);
#else
		value = hub_class_request(gadget, ctrl);
#endif
	} else 
unknown:
	{
		printk(KERN_DEBUG
			"%s: unknown control req%02x.%02x v%04x i%04x l%d\n",
			__FUNCTION__, ctrl->bRequestType, ctrl->bRequest,
			w_value, w_index, w_length);
	}
	
	/* respond with data transfer before status phase? */
	if (respond && value >= 0) {
		req->length = value;
		req->zero = value < w_length
				&& (value % gadget->ep0->maxpacket) == 0;
		value = usb_ep_queue (gadget->ep0, req, GFP_ATOMIC);
		if (value < 0) {
			printk(KERN_DEBUG "%s: ep_queue --> %d\n",
							__FUNCTION__, value);
			req->status = 0;
			hub_setup_complete (gadget->ep0, req);
		}
	}

	/* device either stalls (value < 0) or reports success */
	return value;
}

#ifdef CONFIG_USB_COMPOSITE
static void hub_disconnect (struct usb_composite_dev *cdev)
#else
static void hub_disconnect (struct usb_gadget *gadget)
#endif
{
#ifdef CONFIG_USB_COMPOSITE
	struct hub_dev		*dev = get_composite_data(cdev, function_index);
#else
	struct hub_dev		*dev = get_gadget_data(gadget);
#endif
	unsigned long		flags;

	spin_lock_irqsave (&dev->lock, flags);
	hub_reset_config (dev);
	spin_unlock_irqrestore (&dev->lock, flags);

	/* next we may get setup() calls to enumerate new connections;
	 * or an unbind() during shutdown (including removing module).
	 */
}

static void hub_autoresume (unsigned long _dev)
{
	struct hub_dev	*dev = (struct hub_dev *) _dev;
	int		status;

	/* normally the host would be woken up for something
	 * more significant than just a timer firing...
	 */
	if (dev->gadget->speed != USB_SPEED_UNKNOWN) {
		status = usb_gadget_wakeup (dev->gadget);
		printk(KERN_DEBUG "%s: wakeup --> %d\n", __FUNCTION__, status);
	}
}


/*-------------------------------------------------------------------------*/
#ifdef CONFIG_USB_COMPOSITE
static void hub_unbind (struct usb_composite_dev *cdev)
#else
static void hub_unbind (struct usb_gadget *gadget)
#endif
{
#ifdef CONFIG_USB_COMPOSITE
	struct hub_dev		*dev = get_composite_data(cdev, function_index);
	struct usb_gadget	*gadget = dev->gadget;
#else
	struct hub_dev		*dev = get_gadget_data (gadget);
#endif

	//printk(KERN_DEBUG "%s: unbind\n", __FUNCTION__);
	
	/* Free the request and buffer for endpoint 0 */
	if (dev->req) {
		if (dev->req->buf) {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,23)
			usb_ep_free_buffer(gadget->ep0, dev->req->buf,
						dev->req->dma, USB_BUFSIZ);
#else
			_usb_ep_free_buffer(gadget, gadget->ep0, dev->req->buf,
						dev->req->dma, USB_BUFSIZ);
#endif
		}
		usb_ep_free_request(gadget->ep0, dev->req);
	}
	
	hub_reset_config(dev);

	del_timer_sync (&dev->resume);
	
	if (dev)
		kfree (dev);

#ifdef CONFIG_USB_COMPOSITE
	set_composite_data(cdev, function_index, NULL);
#else
	set_gadget_data(gadget, NULL);
#endif
}

#ifdef CONFIG_USB_COMPOSITE
static int set_string_ids(struct usb_composite_dev *cdev)
{
	struct usb_string *s;
	int               string_id = -EINVAL;
	
	for (s = stringtab.strings; (s && (s->id > 0)); s++)
	{
		string_id = usb_composite_string_id(cdev);
		
		if (string_id < 0)
			break;
		
		switch (s->id)
		{
			case STRING_STATUS_INTERFACE:
				s->id = string_id;
				intf_desc.iInterface = string_id;
				break;
			default:
				return -EINVAL;
		}
	}
	
	return string_id;
}
#endif

#ifdef CONFIG_USB_COMPOSITE
static int __init hub_bind (struct usb_composite_dev *cdev)
#else
static int __init hub_bind (struct usb_gadget *gadget)
#endif
{
	struct hub_dev		*dev;
	int			rc = -EINVAL;
#ifdef CONFIG_USB_COMPOSITE
	struct usb_gadget	*gadget = cdev->gadget;
	int			interface_id;
#endif
	
	dev = kmalloc (sizeof *dev, GFP_KERNEL);
	if (!dev)
		return -ENOMEM;
	memset (dev, 0, sizeof *dev);

	gadget->speed = USB_SPEED_HIGH;
	
	dev->gadget = gadget;
	g_hub = dev;

	/* ok, we made sense of the hardware ... */
	spin_lock_init (&dev->lock);
	
#ifdef CONFIG_USB_COMPOSITE
	// Get function index
	function_index = usb_composite_function_index(cdev);
	if (function_index < 0) {
		printk(KERN_ERR "%s: failed getting function index\n", 
								__FUNCTION__);
		rc = -EINVAL;
		goto out;		
	}	
	set_composite_data(cdev, function_index, dev);
#else
	set_gadget_data (gadget, dev);
#endif

	/* preallocate control response and buffer */
	dev->req = usb_ep_alloc_request (gadget->ep0, GFP_KERNEL);
	if (!dev->req) {
		rc = -ENOMEM;
		goto out;
	}
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,23)
	dev->req->buf = usb_ep_alloc_buffer (gadget->ep0, USB_BUFSIZ,
				&dev->req->dma, GFP_KERNEL);
#else
	dev->req->buf = _usb_ep_alloc_buffer (gadget, gadget->ep0, USB_BUFSIZ,
				&dev->req->dma, GFP_KERNEL);
#endif
	if (!dev->req->buf) {
		rc = -ENOMEM;
		goto out;
	}

	dev->req->complete = hub_setup_complete;
	
#ifdef CONFIG_USB_GADGET_DUALSPEED
	/* and that all endpoints are dual-speed */
	hs_desc.bEndpointAddress = fs_desc.bEndpointAddress;
#endif

#ifdef CONFIG_USB_COMPOSITE
	// Set String IDs
	if (set_string_ids(cdev) < 0) {
		printk(KERN_ERR "%s: failed setting string IDs\n", __FUNCTION__);
		rc = -EINVAL;
		goto out;		
	}
	
	// Set Interface Number for HUB Status
	interface_id = usb_composite_interface_id(cdev, &intf_info);
	if (interface_id >= 0)
		intf_desc.bInterfaceNumber = interface_id;
	else {
		printk(KERN_ERR "%s: invalid Status  interface number %d\n",
						__FUNCTION__, interface_id);
		rc = -EINVAL;
		goto out;
	}

#else
	
	// Set endpoint max packet size in device descriptor
	device_desc.bMaxPacketSize0 = gadget->ep0->maxpacket;

#ifdef CONFIG_USB_GADGET_DUALSPEED
	/* assume ep0 uses the same value for both speeds ... */
	dev_qualifier.bMaxPacketSize0 = device_desc.bMaxPacketSize0;
#endif
#endif

	init_timer (&dev->resume);
	dev->resume.function = hub_autoresume;
	dev->resume.data = (unsigned long) dev;

	gadget->ep0->driver_data = dev;

#ifndef CONFIG_USB_COMPOSITE
	usb_gadget_connect(gadget);
#endif
		
	return 0;

out:
#ifdef CONFIG_USB_COMPOSITE
	hub_unbind (cdev);
#else
	hub_unbind (gadget);
#endif
	return rc;
}

/*-------------------------------------------------------------------------*/

#ifdef CONFIG_USB_COMPOSITE
static void hub_suspend (struct usb_composite_dev *cdev)
#else
static void hub_suspend (struct usb_gadget *gadget)
#endif
{
#ifdef CONFIG_USB_COMPOSITE
	struct hub_dev		*dev = get_composite_data(cdev, function_index);
	struct usb_gadget	*gadget = cdev->gadget;
#else
	struct hub_dev		*dev = get_gadget_data (gadget);
#endif

	if (gadget->speed == USB_SPEED_UNKNOWN)
		return;

	if (autoresume) {
		mod_timer (&dev->resume, jiffies + (HZ * autoresume));
		printk(KERN_DEBUG "%s: suspend, wakeup in %d seconds\n",
						__FUNCTION__, autoresume);
	} else
		printk(KERN_DEBUG "%s: suspend\n", __FUNCTION__);
}

#ifdef CONFIG_USB_COMPOSITE
static void hub_resume (struct usb_composite_dev *cdev)
#else
static void hub_resume (struct usb_gadget *gadget)
#endif
{
#ifdef CONFIG_USB_COMPOSITE
	struct hub_dev		*dev = get_composite_data(cdev, function_index);
#else
	struct hub_dev		*dev = get_gadget_data (gadget);
#endif

	printk(KERN_DEBUG "%s: resume\n", __FUNCTION__);
	del_timer (&(dev->resume));
}


/*-------------------------------------------------------------------------*/
#ifdef CONFIG_USB_COMPOSITE
static struct usb_function hub_driver = {
	.name             = (char *) shortname,
	.controller_name  = controller,
	.strings          = stringTable,
	.descriptors      = fs_hub_function,
#ifdef CONFIG_USB_GADGET_DUALSPEED	
	.hs_descriptors   = hs_hub_function,
#else
	.hs_descriptors   = NULL,
#endif
	.intf_info_table  = intf_info_table,
	.bind             = hub_bind,
	.unbind           = hub_unbind,
	.disconnect       = hub_disconnect,
	.setup            = hub_setup,
	.suspend          = hub_suspend,
	.resume           = hub_resume,
	
	.driver           = {
		.name      = (char *) shortname,
		.owner     = THIS_MODULE,
		// .release = ...
		// .suspend = ...
		// .resume = ...
	},	
};
#else
static struct usb_gadget_driver hub_driver = {
#ifdef CONFIG_USB_GADGET_DUALSPEED
	.speed		= USB_SPEED_HIGH,
#else
	.speed		= USB_SPEED_FULL,
#endif
	.function	= (char *) shortname,
	.bind		= hub_bind,
	.unbind		= hub_unbind,
	.disconnect	= hub_disconnect,
	.setup		= hub_setup,
	.suspend	= hub_suspend,
	.resume		= hub_resume,

	.driver 	= {
		.name		= (char *) shortname,
		.owner     = THIS_MODULE,
		// .release = ...
		// .suspend = ...
		// .resume = ...
	},
};
#endif

MODULE_AUTHOR ("Avocent Corp.");
MODULE_LICENSE ("GPL");

static int __init hub_init (void)
{
	printk(KERN_ALERT "%s: Desc: %s\n", __FUNCTION__, DRIVER_DESC);
	printk(KERN_ALERT "%s: Name: %s\n", __FUNCTION__, DRIVER_NAME);
	printk(KERN_ALERT "%s: Version: %s\n", __FUNCTION__, DRIVER_VERSION);
	
	strcpy(controller, mod_data.controller);	

#ifdef CONFIG_USB_COMPOSITE  	
	return usb_composite_register(&hub_driver);
#else
  #ifndef CONFIG_USB_GADGET_DUAL_CONTROLLERS
	return usb_gadget_register_driver(&hub_driver);
  #endif
#endif 
}
module_init (hub_init);

static void __exit hub_cleanup (void)
{
#ifdef CONFIG_USB_COMPOSITE
	usb_composite_unregister(&hub_driver);
#else
  #ifndef CONFIG_USB_GADGET_DUAL_CONTROLLERS
	usb_gadget_unregister_driver(&hub_driver);
  #endif
#endif	
}
module_exit (hub_cleanup);

